# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A collection of re-usable event handlers for the Copycat UI."""

import dataclasses
import json
import logging

import vertexai
import mesop as me
import pandas as pd

from copycat import copycat
from copycat.data import sheets
from copycat.data import utils as data_utils
from copycat.ui import states


def _force_index_columns_if_possible(
    data: pd.DataFrame,
    index_columns: list[str],
) -> bool:
  """Force the index columns to be set if they exist as columns.

  Sets the index of the data inplace, and returns True if the index columns are
  now correct, False otherwise.

  Args:
    data: The data to set the index columns on.
    index_columns: The columns to set as the index columns.

  Returns:
    True if the index columns are now correct, False otherwise.
  """
  if data.index.names == [None] and set(index_columns).issubset(
      data.columns.values
  ):
    data.set_index(index_columns, inplace=True)
  return data.index.names == index_columns


def read_training_ads(
    sheet: sheets.GoogleSheet, raise_if_bad_index: bool = True
) -> pd.DataFrame:
  """Reads the training ads from the Google Sheet.

  The google sheet should contain a tab named "Training Ads" with the following
  frozen columns: Campaign ID and Ad Group, which are used as the index. If
  there are no frozen columns but the columns Campaign ID and Ad Group exist,
  then they are used as the index. Otherwise, an error is raised.

  Args:
    sheet: The Google Sheet to read the training ads from.
    raise_if_bad_index: If True, then an error is raised if the training ads do
      not have the required index columns.

  Returns:
    The training ads as a pandas DataFrame.

  Raises:
    ValueError: If the training ads do not have the required index columns.
  """
  index_columns = ["Campaign ID", "Ad Group"]
  training_ads = sheet[sheets.TEMPLATE_EXISTING_ADS_WORKSHEET_NAME]
  has_correct_index = _force_index_columns_if_possible(
      training_ads, index_columns
  )
  if not has_correct_index and raise_if_bad_index:
    message = (
        f"Training Ads requires index columns: {index_columns}, but found index"
        f" = {training_ads.index.names} and columns = {training_ads.columns}."
    )
    send_log(message, level=logging.ERROR)
    raise ValueError(message)
  return training_ads


def read_new_keywords(
    sheet: sheets.GoogleSheet, raise_if_bad_index: bool = True
) -> pd.DataFrame:
  """Reads the new keywords from the Google Sheet.

  The google sheet should contain a tab named "New Keywords" with the following
  frozen columns: Campaign ID and Ad Group, which are used as the index. If
  there are no frozen columns but the columns Campaign ID and Ad Group exist,
  then they are used as the index. Otherwise, an error is raised.

  Args:
    sheet: The Google Sheet to read the new keywords from.
    raise_if_bad_index: If True, then an error is raised if the new keywords do
      not have the required index columns.

  Returns:
    The new keywords as a pandas DataFrame.

  Raises:
    ValueError: If the new keywords do not have the required index columns.
  """
  index_columns = ["Campaign ID", "Ad Group"]
  new_keywords = sheet[sheets.TEMPLATE_NEW_KEYWORDS_WORKSHEET_NAME]
  has_correct_index = _force_index_columns_if_possible(
      new_keywords, index_columns
  )
  if not has_correct_index and raise_if_bad_index:
    message = (
        f"New Keywords requires index columns: {index_columns}, but found index"
        f" = {new_keywords.index.names} and columns = {new_keywords.columns}."
    )
    send_log(message, level=logging.ERROR)
    raise ValueError(message)
  return new_keywords


def read_extra_instructions(
    sheet: sheets.GoogleSheet, raise_if_bad_index: bool = True
) -> pd.DataFrame:
  """Reads the extra instructions from the Google Sheet.

  The google sheet should contain a tab named "Extra Instructions for New Ads"
  with the following frozen columns: Campaign ID, Ad Group and Version, which
  are used as the index. If there are no frozen columns but the columns Campaign
  ID, Ad Group and Version exist, then they are used as the index. Otherwise, an
  error is raised.

  Args:
    sheet: The Google Sheet to read the extra instructions from.
    raise_if_bad_index: If True, then an error is raised if the extra
      instructions do not have the required index columns.

  Returns:
    The extra instructions as a pandas DataFrame.

  Raises:
    ValueError: If the extra instructions do not have the required index
    columns.
  """
  index_columns = ["Campaign ID", "Ad Group", "Version"]
  extra_instructions = sheet[sheets.TEMPLATE_EXTRA_INSTRUCTIONS_WORKSHEET_NAME]
  has_correct_index = _force_index_columns_if_possible(
      extra_instructions, index_columns
  )
  if not has_correct_index and raise_if_bad_index:
    message = (
        f"Extra Instructions requires index columns: {index_columns}, but found"
        f" index = {extra_instructions.index.names} and columns ="
        f" {extra_instructions.columns}."
    )
    send_log(message, level=logging.ERROR)
    raise ValueError(message)
  return extra_instructions


def read_generated_ads(
    sheet: sheets.GoogleSheet, raise_if_bad_index: bool = True
) -> pd.DataFrame:
  """Reads the generated ads from the Google Sheet.

  The google sheet should contain a tab named "Generated Ads"
  with the following frozen columns: Campaign ID, Ad Group and Version, which
  are used as the index. If there are no frozen columns but the columns Campaign
  ID, Ad Group and Version exist, then they are used as the index. Otherwise, an
  error is raised.

  Args:
    sheet: The Google Sheet to read the generated ads from.
    raise_if_bad_index: If True, then an error is raised if the generated ads do
      not have the required index columns.

  Returns:
    The generated ads as a pandas DataFrame.

  Raises:
    ValueError: If the generated ads do not have the required index columns.
  """
  index_columns = ["Campaign ID", "Ad Group", "Version"]
  generated_ads = sheet["Generated Ads"]
  has_correct_index = _force_index_columns_if_possible(
      generated_ads, index_columns
  )
  if not has_correct_index and raise_if_bad_index:
    message = (
        f"Extra Instructions requires index columns: {index_columns}, but found"
        f" index = {generated_ads.index.names} and columnms ="
        f" {generated_ads.columns}."
    )
    send_log(message, level=logging.ERROR)
    raise ValueError(message)
  return generated_ads


def update_copycat_parameter(event: me.InputEvent) -> None:
  """Updates a parameter in the CopycatParamsState.

  Args:
    event: The input event to handle. This can be any event where the key is
      set.

  Raises:
    ValueError: If the key is not a field in CopycatParamsState.
  """
  params = me.state(states.CopycatParamsState)
  for field in dataclasses.fields(params):
    if field.name == event.key:
      setattr(params, event.key, field.type(event.value))
      return
  raise ValueError(f"Field {event.key} does not exist in CopycatParamsState.")


def update_app_state_parameter(event: me.InputEvent) -> None:
  """Updates a parameter in the AppState.

  Args:
    event: The input event to handle. This can be any event where the key is
      set.

  Raises:
    ValueError: If the key is not a field in AppState.
  """
  state = me.state(states.AppState)
  for field in dataclasses.fields(state):
    if field.name == event.key:
      setattr(state, event.key, field.type(event.value))
      return
  raise ValueError(f"Field {event.key} does not exist in AppState.")


def update_copycat_parameter_from_slide_toggle(
    event: me.SlideToggleChangeEvent,
) -> None:
  """Updates a copycat parameter from a slide toggle change event.

  Args:
    event: The slide toggle change event to handle.
  """
  state = me.state(states.CopycatParamsState)
  setattr(state, event.key, not getattr(state, event.key))


def update_app_state_parameter_checkbox(
    event: me.CheckboxChangeEvent,
) -> None:
  """Updates a app state parameter from a checkbox change event.

  Args:
    event: The checkbox change event to handle.
  """
  state = me.state(states.AppState)
  setattr(state, event.key, not getattr(state, event.key))


def language_on_blur(event: me.InputBlurEvent) -> None:
  """Updates the language and the embedding model name based on the language.

  Args:
    event: The input blur event to handle.
  """
  state = me.state(states.CopycatParamsState)
  state.language = event.value

  if "english" in event.value.lower():
    state.embedding_model_name = copycat.EmbeddingModelName.TEXT_EMBEDDING.value
  else:
    state.embedding_model_name = (
        copycat.EmbeddingModelName.TEXT_MULTILINGUAL_EMBEDDING.value
    )

  send_log(
      f"Updating embedding model name to {state.embedding_model_name} for"
      f" language = {state.language}"
  )


def ad_format_on_change(event: me.RadioChangeEvent) -> None:
  """Updates the ad format and related parameters in the CopycatParamsState.

  Args:
    event: The radio change event to handle.
  """
  state = me.state(states.CopycatParamsState)
  state.ad_format = event.value

  if state.ad_format != "custom":
    send_log(
        f"Updating max headlines and descriptions for {state.ad_format} ad"
        " format"
    )
    ad_format = copycat.google_ads.get_google_ad_format(event.value)
    state.max_headlines = ad_format.max_headlines
    state.max_descriptions = ad_format.max_descriptions
  else:
    send_log(
        f"Max headlines and descriptions not updated for {state.ad_format} ad"
        " format"
    )


def embedding_model_dimensionality_on_blur(event: me.InputBlurEvent) -> None:
  """Updates the embedding model dimensionality based on the input value.

  The value is clamped to the range [10, 768].

  Args:
    event: The input blur event to handle.
  """
  state = me.state(states.CopycatParamsState)
  raw_value = int(event.value)
  if raw_value > 786:
    state.embedding_model_dimensionality = 768
  elif raw_value < 10:
    state.embedding_model_dimensionality = 10
  else:
    state.embedding_model_dimensionality = raw_value


def close_starting_dialog(event: me.ClickEvent) -> None:
  """Closes the starting dialog.

  This clears the new Google Sheet URL and name, and sets the
  show_starting_dialog state to False.

  Args:
    event: The click event to handle.
  """
  state = me.state(states.AppState)
  state.new_google_sheet_url = ""
  state.new_google_sheet_name = ""
  state.show_starting_dialog = False


def open_starting_dialog(event: me.ClickEvent) -> None:
  """Opens the starting dialog by setting show_starting_dialog to True.

  Args:
    event: The click event to handle.
  """
  state = me.state(states.AppState)
  state.show_starting_dialog = True


def reset_state(
    state: type[states.AppState] | type[states.CopycatParamsState],
) -> None:
  """Resets a state to its default values.

  Args:
    state: The state to reset.
  """
  send_log(f"Resetting state: {state}")
  params = me.state(state)

  for field in dataclasses.fields(params):
    if field.default is not dataclasses.MISSING:
      setattr(params, field.name, field.default)
    elif field.default_factory is not dataclasses.MISSING:
      setattr(params, field.name, field.default_factory())
    else:
      setattr(params, field.name, field.type())


def save_params_to_google_sheet(event: me.ClickEvent) -> None:
  """Saves the Copycat parameters to the Google Sheet.

  The parameters are written to a tab named "READ ONLY: Copycat Params".

  Args:
    event: The click event to handle.
  """
  state = me.state(states.AppState)
  params = me.state(states.CopycatParamsState)

  params_table = pd.DataFrame([dataclasses.asdict(params)])
  params_table["Parameter Name"] = "Parameter Value"
  params_table = params_table.set_index("Parameter Name")

  sheet = sheets.GoogleSheet.load(state.google_sheet_url)
  sheet["READ ONLY: Copycat Params"] = params_table

  send_log("Copycat params saved to sheet")


def load_params_from_google_sheet(event: me.ClickEvent) -> None:
  """Loads the Copycat parameters from the Google Sheet.

  The parameters are read from a tab named "READ ONLY: Copycat Params".

  Args:
    event: The click event to handle.
  """
  state = me.state(states.AppState)
  params = me.state(states.CopycatParamsState)

  sheet = sheets.GoogleSheet.load(state.google_sheet_url)
  params_table = sheet["READ ONLY: Copycat Params"]

  for field in dataclasses.fields(params):
    if field.name in params_table:
      param_value = params_table[field.name].values[0]
      if field.type is bool and param_value == "TRUE":
        param_value = True
      elif field.type is bool and param_value == "FALSE":
        param_value = False
      else:
        param_value = field.type(param_value)
      setattr(params, field.name, param_value)

  state.has_copycat_instance = "READ ONLY: Copycat Instance Params" in sheet
  send_log("Loaded Copycat params from sheet")


def create_new_google_sheet(event: me.ClickEvent) -> None:
  """Creates a new Google Sheet and initializes it with the default tabs.

  The default tabs are:
    - Training Ads
    - New Keywords
    - Extra Instructions for New Ads

  Args:
    event: The click event to handle.
  """
  state = me.state(states.AppState)
  new_sheet_url = sheets.create_template_copycat_sheet(
      state.new_google_sheet_name, state.new_google_sheet_include_demo_data
  )
  start_logger(new_sheet_url)
  send_log(
      f"Created new Google Sheet: {state.new_google_sheet_name}. Include demo"
      f" data: {state.new_google_sheet_include_demo_data}."
  )

  reset_state(states.AppState)
  reset_state(states.CopycatParamsState)
  state = me.state(states.AppState)

  sheet = sheets.GoogleSheet.load(new_sheet_url)
  state.google_sheet_url = sheet.url
  state.google_sheet_name = sheet.title
  save_params_to_google_sheet(event)
  close_starting_dialog(event)
  send_log("New Google Sheet created")


def load_existing_google_sheet(event: me.ClickEvent) -> None:
  """Loads an existing Google Sheet.

  The sheet should contain the following tabs:
    - Training Ads
    - New Keywords
    - Extra Instructions for New Ads

  Args:
    event: The click event to handle.
  """
  state = me.state(states.AppState)
  sheet = sheets.GoogleSheet.load(state.new_google_sheet_url)
  start_logger(sheet.url)

  reset_state(states.AppState)
  reset_state(states.CopycatParamsState)
  state = me.state(states.AppState)

  # Load sheet
  state.google_sheet_url = sheet.url
  state.google_sheet_name = sheet.title

  if "READ ONLY: Copycat Params" in sheet:
    load_params_from_google_sheet(event)
  else:
    save_params_to_google_sheet(event)

  close_starting_dialog(event)
  send_log("Existing Google Sheet loaded")


def start_logger(url: str) -> None:
  """Starts the logger and writes logs to a Google Sheet.

  Args:
    url: The URL of the Google Sheet to write logs to.
  """
  handler = sheets.GoogleSheetsHandler(sheet_url=url, log_worksheet_name="Logs")
  handler.setLevel(logging.INFO)


  logger = logging.getLogger("copycat")
  logger.handlers = []
  logger.addHandler(handler)
  logger.setLevel(logging.INFO)
  logger.info("Logger Started")


def send_log(message: str, level: int = logging.INFO) -> None:
  """Sends a log message to the logger.

  Args:
    message: The log message to send.
    level: The level of the log message. Defaults to INFO.
  """
  logger = logging.getLogger("copycat.ui")
  logger.log(level=level, msg=message)


def update_log_level(event: me.SelectSelectionChangeEvent) -> None:
  """Updates the log level of the logger.

  Args:
    event: The select selection change event to handle.
  """
  state = me.state(states.AppState)
  state.log_level = int(event.value)

  loggers = [
      logging.getLogger(name) for name in logging.root.manager.loggerDict
  ]
  for logger in loggers:
    if "copycat" in logger.name:
      logger.setLevel(state.log_level)
      for handler in logger.handlers:
        if isinstance(handler, sheets.GoogleSheetsHandler):
          handler.setLevel(state.log_level)


def show_hide_google_sheet(event: me.ClickEvent) -> None:
  """Shows or hides the Google Sheet preview panel.

  Args:
    event: The click event to handle.
  """
  state = me.state(states.AppState)
  state.display_google_sheet = not state.display_google_sheet


def on_click_snackbar_close(event: me.ClickEvent):
  state = me.state(states.AppState)
  setattr(state, event.key, False)


def validate_sheet(event: me.ClickEvent) -> None:
  """Validates the Google Sheet.

  The sheet is validated by checking that it contains the required tabs,
  index columns, and columns, and that it has the minimum number of rows.

  Args:
    event: The click event to handle.
  """
  state = me.state(states.AppState)
  sheet_url = state.google_sheet_url
  send_log(f"Validating {sheet_url}")

  sheet = sheets.GoogleSheet.load(sheet_url)
  send_log(f"Sheet Name = {sheet.title}")

  # Validate all required sheets exist, have the correct index and columns,
  # and have the minimum number of rows.
  required_index_names = {
      sheets.TEMPLATE_EXISTING_ADS_WORKSHEET_NAME: ["Campaign ID", "Ad Group"],
      sheets.TEMPLATE_NEW_KEYWORDS_WORKSHEET_NAME: ["Campaign ID", "Ad Group"],
      sheets.TEMPLATE_EXTRA_INSTRUCTIONS_WORKSHEET_NAME: [
          "Campaign ID",
          "Ad Group",
          "Version",
      ],
  }
  required_columns = {
      sheets.TEMPLATE_EXISTING_ADS_WORKSHEET_NAME: set([
          "URL",
          "Ad Strength",
          "Keywords",
          "Headline 1",
          "Description 1",
      ]),
      sheets.TEMPLATE_NEW_KEYWORDS_WORKSHEET_NAME: set([
          "Keyword",
      ]),
      sheets.TEMPLATE_EXTRA_INSTRUCTIONS_WORKSHEET_NAME: set([
          "Extra Instructions",
      ]),
  }
  min_rows = {
      sheets.TEMPLATE_EXISTING_ADS_WORKSHEET_NAME: 1,
      sheets.TEMPLATE_NEW_KEYWORDS_WORKSHEET_NAME: 1,
      sheets.TEMPLATE_EXTRA_INSTRUCTIONS_WORKSHEET_NAME: 0,
  }

  state.google_sheet_is_valid = True
  for sheet_name in required_columns:
    if sheet_name in sheet:
      send_log(f"{sheet_name} sheet found")
    else:
      send_log(
          f"VALIDATION FAILED: {sheet_name} sheet not found.", logging.ERROR
      )
      state.google_sheet_is_valid = False
      continue

    worksheet = sheet[sheet_name]
    _force_index_columns_if_possible(
        worksheet, required_index_names[sheet_name]
    )
    actual_index_names = list(worksheet.index.names)
    if required_index_names[sheet_name] != actual_index_names:
      send_log(
          f"VALIDATION FAILED: {sheet_name} requires index columns:"
          f" {required_index_names[sheet_name]}, but found"
          f" {actual_index_names}.",
          logging.ERROR,
      )
      state.google_sheet_is_valid = False

    actual_columns = set(worksheet.columns.values.tolist())
    extra_columns = actual_columns - required_columns[sheet_name]
    missing_columns = required_columns[sheet_name] - actual_columns

    if missing_columns:
      send_log(
          f"VALIDATION FAILED: Missing columns in {sheet_name}:"
          f" {missing_columns}",
          logging.ERROR,
      )
      state.google_sheet_is_valid = False
    else:
      send_log(f"All required columns in {sheet_name}")

    if extra_columns:
      send_log(f"{sheet_name} has the following extra columns: {extra_columns}")

    n_rows = len(worksheet)
    if n_rows < min_rows[sheet_name]:
      send_log(
          f"VALIDATION FAILED: {sheet_name} sheet has fewer than the minimum"
          f" number of rows: min={min_rows[sheet_name]}.",
          logging.ERROR,
      )
      state.google_sheet_is_valid = False
    else:
      send_log(f"{sheet_name} has {n_rows:,} rows")

  # Log the number of headline and description columns in the training ads
  training_ads = read_training_ads(sheet, raise_if_bad_index=False)
  n_headline_columns = len(
      [c for c in training_ads.columns if c.startswith("Headline")]
  )
  n_description_columns = len(
      [c for c in training_ads.columns if c.startswith("Description")]
  )
  send_log(f"Training Ads have up to {n_headline_columns} headlines.")
  send_log(f"Training Ads have up to {n_description_columns} descriptions.")

  # Completed validation
  if state.google_sheet_is_valid:
    send_log("VALIDATION COMPLETED: Google Sheet is valid")
  else:
    send_log("VALIDATION COMPLETED: Google Sheet is invalid", logging.ERROR)


def save_copycat_to_sheet(
    sheet: sheets.GoogleSheet, model: copycat.Copycat
) -> None:
  """Saves the Copycat model to the Google Sheet.

  The model is saved in two tabs:
    - READ ONLY: Training Ad Exemplars: Contains the exemplar ads.
    - READ ONLY: Copycat Instance Params: Contains the other model parameters.

  Args:
    sheet: The Google Sheet to save the model to.
    model: The Copycat model to save.
  """
  send_log("Saving Copycat instance to sheet")
  model_params = model.to_dict()

  # Store the exemplar ads in their own sheet
  exemplars_dict = model_params["ad_copy_vectorstore"].pop("ad_exemplars")
  ad_exemplars = pd.DataFrame.from_dict(exemplars_dict, orient="tight")
  ad_exemplars["embeddings"] = ad_exemplars["embeddings"].apply(
      lambda x: ", ".join(list(map(str, x)))
  )
  ad_exemplars = data_utils.explode_headlines_and_descriptions(ad_exemplars)
  ad_exemplars.index.name = "Exemplar Number"
  sheet["READ ONLY: Training Ad Exemplars"] = ad_exemplars

  # Store the other params as a json string
  other_params = pd.DataFrame([{
      "params_json": json.dumps(model_params),
  }])
  sheet["READ ONLY: Copycat Instance Params"] = other_params


def load_copycat_from_sheet(sheet: sheets.GoogleSheet) -> copycat.Copycat:
  """Loads a Copycat instance from the Google Sheet.

  The instance is loaded from two tabs:
    - READ ONLY: Training Ad Exemplars: Contains the exemplar ads.
    - READ ONLY: Copycat Instance Params: Contains the other model parameters.

  Args:
    sheet: The Google Sheet to load the instance from.

  Returns:
    The Copycat instance.
  """
  send_log("Loading Copycat instance from sheet")
  instance_json = sheet["READ ONLY: Copycat Instance Params"].loc[
      0, "params_json"
  ]
  instance_dict = json.loads(instance_json)

  ad_exemplars = sheet["READ ONLY: Training Ad Exemplars"]
  ad_exemplars["embeddings"] = ad_exemplars["embeddings"].apply(
      lambda x: list(map(float, x.split(", ")))
  )
  ad_exemplars = data_utils.collapse_headlines_and_descriptions(ad_exemplars)
  ad_exemplars_dict = ad_exemplars.to_dict(orient="tight")

  instance_dict["ad_copy_vectorstore"]["ad_exemplars"] = ad_exemplars_dict
  copycat_instance = copycat.Copycat.from_dict(instance_dict)
  return copycat_instance


def build_new_copycat_instance(event: me.ClickEvent):
  """Builds a new Copycat instance from the Google Sheet.

  The Copycat instance is created using the parameters from the Google Sheet
  and the CopycatParamsState. The instance is then saved to the Google Sheet.

  Args:
    event: The click event to handle.
  """
  state = me.state(states.AppState)
  params = me.state(states.CopycatParamsState)
  sheet = sheets.GoogleSheet.load(state.google_sheet_url)
  save_params_to_google_sheet(event)

  vertexai.init(
      project=params.vertex_ai_project_id, location=params.vertex_ai_location
  )

  train_data = data_utils.collapse_headlines_and_descriptions(
      read_training_ads(sheet)
  )
  train_data = train_data.rename({"Keywords": "keywords"}, axis=1)
  train_data = train_data[["headlines", "descriptions", "keywords"]]
  train_data = train_data.loc[train_data["headlines"].apply(len) > 0]

  send_log(f"Loaded {len(train_data)} rows of raw data from the Google Sheet.")

  if params.ad_format == "custom":
    ad_format = copycat.google_ads.GoogleAdFormat(
        name="custom",
        max_headlines=params.max_headlines,
        max_descriptions=params.max_descriptions,
        min_headlines=1,
        min_descriptions=1,
        max_headline_length=30,
        max_description_length=90,
    )
    send_log("Using a custom ad format.")
  else:
    ad_format = copycat.google_ads.get_google_ad_format(params.ad_format)
    send_log(f"Using the following ad format: {ad_format.name}")

  affinity_preference = (
      params.custom_affinity_preference
      if params.use_custom_affinity_preference
      else None
  )
  send_log(
      "Affinity preference:"
      f" {affinity_preference} (custom={params.use_custom_affinity_preference})"
  )

  send_log("Creating Copycat.")

  model = copycat.Copycat.create_from_pandas(
      training_data=train_data,
      ad_format=ad_format,
      on_invalid_ad=params.on_invalid_ad,
      embedding_model_name=params.embedding_model_name,
      embedding_model_dimensionality=params.embedding_model_dimensionality,
      embedding_model_batch_size=params.embedding_model_batch_size,
      vectorstore_exemplar_selection_method=params.exemplar_selection_method,
      vectorstore_max_initial_ads=params.max_initial_ads,
      vectorstore_max_exemplar_ads=params.max_exemplar_ads,
      vectorstore_affinity_preference=affinity_preference,
      replace_special_variables_with_default=params.how_to_handle_special_variables
      == "replace",
  )
  send_log(
      "Copycat instance created with"
      f" {model.ad_copy_vectorstore.n_exemplars} exemplar ads."
  )

  save_copycat_to_sheet(sheet, model)
  state.has_copycat_instance = True

  send_log("Copycat instance stored in google sheet.")
  state.show_copycat_instance_created_snackbar = True


def generate_style_guide(event: me.ClickEvent):
  """Generates a style guide from the Google Sheet.

  The style guide is generated using the parameters from the Google Sheet
  and the CopycatParamsState. The style guide is then saved to the Google Sheet.

  Args:
    event: The click event to handle.
  """
  send_log("Generating style guide")
  state = me.state(states.AppState)
  params = me.state(states.CopycatParamsState)
  sheet = sheets.GoogleSheet.load(state.google_sheet_url)

  vertexai.init(
      project=params.vertex_ai_project_id,
      location=params.vertex_ai_location,
  )

  send_log("Preparing to generate style guide")
  if params.style_guide_files_uri:
    send_log(
        f"Checking for files in the GCP bucket {params.style_guide_files_uri}"
    )

  copycat_instance = load_copycat_from_sheet(sheet)
  params.style_guide = copycat_instance.generate_style_guide(
      company_name=params.company_name,
      additional_style_instructions=params.style_guide_additional_instructions,
      model_name=params.style_guide_chat_model_name,
      safety_settings=copycat.ALL_SAFETY_SETTINGS_ONLY_HIGH,
      temperature=params.style_guide_temperature,
      top_k=params.style_guide_top_k,
      top_p=params.style_guide_top_p,
      use_exemplar_ads=params.style_guide_use_exemplar_ads,
      files_uri=params.style_guide_files_uri,
  )
  send_log("Style guide generated")

  save_params_to_google_sheet(event)


def _prepare_new_ads_for_generation(
    sheet: sheets.GoogleSheet,
    n_versions: int,
    fill_gaps: bool,
    copycat_instance: copycat.Copycat,
) -> tuple[pd.DataFrame, pd.DataFrame]:
  """Prepares new ads for generation.

  This function prepares new ads for generation by loading the data from the
  Google Sheet, constructing the complete data, and filtering out any ads that
  have already been generated.

  Args:
    sheet: The Google Sheet to load the data from.
    n_versions: The number of versions to generate for each ad.
    fill_gaps: Whether to fill gaps in the existing generations.
    copycat_instance: The Copycat instance to use for generation.

  Returns:
    A tuple containing the new generations data and the complete data.
  """
  new_keywords_data = read_new_keywords(sheet)
  additional_instructions_data = read_extra_instructions(sheet)
  additional_instructions_data.index = (
      additional_instructions_data.index.set_levels(
          additional_instructions_data.index.get_level_values("Version").astype(
              str
          ),
          level="Version",
          verify_integrity=False,
      )
  )

  if "Generated Ads" in sheet:
    existing_generations_data = read_generated_ads(sheet)
    if "Headline 1" not in existing_generations_data.columns:
      existing_generations_data = None
  else:
    existing_generations_data = None

  if existing_generations_data is None:
    complete_data = data_utils.construct_generation_data(
        new_keywords_data=new_keywords_data,
        additional_instructions_data=additional_instructions_data,
        n_versions=n_versions,
        keyword_column="Keyword",
        version_column="Version",
        additional_instructions_column="Extra Instructions",
    )
    complete_data["existing_headlines"] = [[]] * len(complete_data)
    complete_data["existing_descriptions"] = [[]] * len(complete_data)
    new_generations_data = complete_data.copy()
    return new_generations_data, complete_data

  existing_generations_data = data_utils.collapse_headlines_and_descriptions(
      existing_generations_data
  ).rename(
      columns={
          "headlines": "existing_headlines",
          "descriptions": "existing_descriptions",
      }
  )
  existing_generations_data.index = existing_generations_data.index.set_levels(
      existing_generations_data.index.get_level_values("Version").astype(str),
      level="Version",
      verify_integrity=False,
  )

  complete_data = data_utils.construct_generation_data(
      new_keywords_data=new_keywords_data,
      additional_instructions_data=additional_instructions_data,
      existing_generations_data=existing_generations_data,
      n_versions=n_versions,
      existing_headlines_column="existing_headlines",
      existing_descriptions_column="existing_descriptions",
      keyword_column="Keyword",
      version_column="Version",
      additional_instructions_column="Extra Instructions",
  )
  missing_columns = [
      column
      for column in existing_generations_data.columns
      if column not in complete_data.columns
  ]
  if missing_columns:
    complete_data = complete_data.join(
        existing_generations_data[missing_columns],
        how="left",
    )

  generation_not_required = complete_data.index.isin(
      existing_generations_data.index
  )
  if fill_gaps:
    generation_not_required = generation_not_required & complete_data.apply(
        lambda row: copycat_instance.ad_copy_evaluator.is_complete(
            copycat.GoogleAd(
                headlines=row["existing_headlines"],
                descriptions=row["existing_descriptions"],
            )
        ),
        axis=1,
    )
  else:
    generation_not_required = generation_not_required & complete_data.apply(
        lambda row: not copycat_instance.ad_copy_evaluator.is_empty(
            copycat.GoogleAd(
                headlines=row["existing_headlines"],
                descriptions=row["existing_descriptions"],
            )
        ),
        axis=1,
    )

  new_generations_data = complete_data.loc[~generation_not_required].copy()

  return new_generations_data, complete_data


def generate_new_ad_preview(event: me.ClickEvent):
  """Generates a preview of the prompt to be used to generate a new ad.

  The preview is generated using the parameters from the Google Sheet
  and the CopycatParamsState. The preview is then saved to the Google Sheet.

  Args:
    event: The click event to handle.
  """
  state = me.state(states.AppState)
  params = me.state(states.CopycatParamsState)
  sheet = sheets.GoogleSheet.load(state.google_sheet_url)
  copycat_instance = load_copycat_from_sheet(sheet)

  generation_data, complete_data = _prepare_new_ads_for_generation(
      sheet,
      params.new_ads_number_of_versions,
      params.new_ads_fill_gaps,
      copycat_instance,
  )

  if len(generation_data) > 0:
    first_row = generation_data.iloc[0].to_dict()
  elif len(complete_data) > 0:
    send_log(
        "No ads to generate, generating a preview for an existing ad",
        logging.WARNING,
    )
    first_row = complete_data.iloc[0].to_dict()
  else:
    send_log(
        "No new or existing ads to generate, no preview will be generated",
        logging.ERROR,
    )
    return

  vertexai.init(
      project=params.vertex_ai_project_id,
      location=params.vertex_ai_location,
  )

  style_guide = params.style_guide if params.new_ads_use_style_guide else ""
  headlines = (
      [first_row["existing_headlines"]]
      if "existing_headlines" in first_row
      else [[]]
  )
  descriptions = (
      [first_row["existing_descriptions"]]
      if "existing_descriptions" in first_row
      else [[]]
  )

  request = copycat_instance.construct_text_generation_requests_for_new_ad_copy(
      keywords=[first_row["keywords"]],
      keywords_specific_instructions=[first_row["additional_instructions"]],
      system_instruction_kwargs=dict(
          company_name=params.company_name,
          language=params.language,
      ),
      style_guide=style_guide,
      num_in_context_examples=params.new_ads_num_in_context_examples,
      model_name=params.new_ads_chat_model_name,
      temperature=params.new_ads_temperature,
      top_k=params.new_ads_top_k,
      top_p=params.new_ads_top_p,
      safety_settings=copycat.ALL_SAFETY_SETTINGS_ONLY_HIGH,
      existing_headlines=headlines,
      existing_descriptions=descriptions,
  )[0]

  state.new_ad_preview_request = request.to_markdown()
  save_params_to_google_sheet(event)


def generate_ads(event: me.ClickEvent):
  """Generates ads from the Google Sheet.

  The ads are generated using the parameters from the Google Sheet
  and the CopycatParamsState. The ads are then saved to the Google Sheet.

  The ads are generated in batches and the batches are written immediately to
  the Google Sheet once they are generated.

  Args:
    event: The click event to handle.
  """
  state = me.state(states.AppState)
  params = me.state(states.CopycatParamsState)
  sheet = sheets.GoogleSheet.load(state.google_sheet_url)
  copycat_instance = load_copycat_from_sheet(sheet)

  save_params_to_google_sheet(event)

  vertexai.init(
      project=params.vertex_ai_project_id,
      location=params.vertex_ai_location,
  )

  generation_data, complete_data = _prepare_new_ads_for_generation(
      sheet,
      params.new_ads_number_of_versions,
      params.new_ads_fill_gaps,
      copycat_instance,
  )
  updated_complete_data = data_utils.explode_headlines_and_descriptions(
      complete_data.copy().rename(
          columns={
              "existing_headlines": "headlines",
              "existing_descriptions": "descriptions",
          }
      ),
      max_headlines=params.max_headlines,
      max_descriptions=params.max_descriptions,
  )

  send_log("Loaded generation and complete data")

  if len(generation_data) == 0:
    send_log("No ads to generate", logging.WARNING)
    return

  generation_params = dict(
      system_instruction_kwargs=dict(
          company_name=params.company_name,
          language=params.language,
      ),
      num_in_context_examples=params.new_ads_num_in_context_examples,
      model_name=params.new_ads_chat_model_name,
      temperature=params.new_ads_temperature,
      top_k=params.new_ads_top_k,
      top_p=params.new_ads_top_p,
      allow_memorised_headlines=params.new_ads_allow_memorised_headlines,
      allow_memorised_descriptions=params.new_ads_allow_memorised_descriptions,
      safety_settings=copycat.ALL_SAFETY_SETTINGS_ONLY_HIGH,
      style_guide=params.style_guide if params.new_ads_use_style_guide else "",
  )
  limit = params.new_ads_generation_limit
  if limit == 0:
    limit = None
  data_iterator = data_utils.iterate_over_batches(
      generation_data,
      batch_size=params.new_ads_batch_size,
      limit_rows=limit,
  )
  for batch_number, generation_batch in enumerate(data_iterator):
    send_log(f"Generating batch {batch_number+1}")
    generation_batch["generated_ad_object"] = (
        copycat_instance.generate_new_ad_copy_for_dataframe(
            data=generation_batch,
            keywords_specific_instructions_column="additional_instructions",
            **generation_params,
        )
    )
    generation_batch = (
        generation_batch.pipe(data_utils.explode_generated_ad_object)
        .pipe(
            data_utils.explode_headlines_and_descriptions,
            max_headlines=params.max_headlines,
            max_descriptions=params.max_descriptions,
        )
        .drop(
            columns=[
                "generated_ad_object",
                "existing_headlines",
                "existing_descriptions",
            ],
            errors="ignore",
        )
    )

    isin_batch = updated_complete_data.index.isin(generation_batch.index)

    updated_complete_data = updated_complete_data.loc[~isin_batch]
    updated_complete_data = pd.concat([updated_complete_data, generation_batch])
    updated_complete_data = updated_complete_data.fillna("").loc[
        complete_data.index
    ]

    column_order = ["keywords", "additional_instructions"]
    column_order.extend(
        col
        for col in updated_complete_data.columns
        if col.startswith("Headline ")
        and col.split(" ")[1].isdigit()
        and len(col.split(" ")) == 2
    )
    column_order.extend(
        col
        for col in updated_complete_data.columns
        if col.startswith("Description ")
        and col.split(" ")[1].isdigit()
        and len(col.split(" ")) == 2
    )
    column_order.extend(
        col for col in updated_complete_data.columns if col not in column_order
    )

    sheet["Generated Ads"] = updated_complete_data[column_order]

  send_log("Generation Complete")
  state.show_ad_copy_generated_snackbar = True



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
  from copycat.ui import main
  from copycat.ui import states

  set_default_gcp_project_id = states.set_default_gcp_project_id
except ImportError:
  # The UI may not be available if the user has not installed Mesop. In this
  # case, we will just ignore the import error.
  print("Copycat UI is not available. Please install Mesop to use the UI.")
  pass



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mesop as me

from copycat import copycat
from copycat.ui import components
from copycat.ui import event_handlers
from copycat.ui import states
from copycat.ui import styles


_STYLE_GUIDE_EDITOR_STYLE = me.Style(
    overflow_y="scroll",
    padding=me.Padding(left=20, right=15, top=20, bottom=0),
    border=me.Border(
        right=styles.DEFAULT_BORDER_STYLE,
    ),
    width="50%",
)

_STYLE_GUIDE_PREVIEW_STYLE = me.Style(
    overflow_y="scroll",
    padding=me.Padding.symmetric(vertical=0, horizontal=20),
    width="50%",
)

_STYLE_GUIDE_TEXTAREA_STYLE = me.Style(
    outline="none",  # Hides focus border
    border=me.Border.all(me.BorderSide(style="none")),
    width="100%",
    height="100%",
    background=me.theme_var("background"),
)


def style_guide():
  """Renders the style guide page."""
  params = me.state(states.CopycatParamsState)

  with components.column(width="100%"):
    with components.row(width="100%"):
      with components.rounded_box_section("Generate Style Guide", width="100%"):
        me.text("Chat Model:")
        me.radio(
            on_change=event_handlers.update_copycat_parameter,
            key="style_guide_chat_model_name",
            options=[
                me.RadioOption(
                    label="Gemini 1.5 Flash",
                    value=copycat.ModelName.GEMINI_1_5_FLASH.value,
                ),
                me.RadioOption(
                    label="Gemini 1.5 Pro",
                    value=copycat.ModelName.GEMINI_1_5_PRO.value,
                ),
            ],
            value=params.style_guide_chat_model_name,
            style=me.Style(margin=me.Margin(bottom=15)),
        )
        with components.row():
          me.input(
              label="Temperature",
              on_blur=event_handlers.update_copycat_parameter,
              key="style_guide_temperature",
              value=str(params.style_guide_temperature),
              appearance="outline",
              type="number",
              style=me.Style(margin=me.Margin.symmetric(horizontal=5)),
          )
          me.input(
              label="Top-K",
              on_blur=event_handlers.update_copycat_parameter,
              key="style_guide_top_k",
              value=str(params.style_guide_top_k),
              appearance="outline",
              type="number",
              style=me.Style(margin=me.Margin.symmetric(horizontal=5)),
          )
          me.input(
              label="Top-P",
              on_blur=event_handlers.update_copycat_parameter,
              key="style_guide_top_p",
              value=str(params.style_guide_top_p),
              appearance="outline",
              type="number",
              style=me.Style(margin=me.Margin.symmetric(horizontal=5)),
          )
        me.slide_toggle(
            label="Use exemplar ads to generate Style Guide",
            on_change=event_handlers.update_copycat_parameter_from_slide_toggle,
            key="style_guide_use_exemplar_ads",
            checked=params.style_guide_use_exemplar_ads,
        )
        me.text(
            "Additional instructions for generating the style guide"
            " (optional):",
            style=me.Style(margin=me.Margin.symmetric(vertical=15)),
        )
        with me.box(style=me.Style(width="40%")):
          me.input(
              label="Additional Instructions",
              on_blur=event_handlers.update_copycat_parameter,
              key="style_guide_additional_instructions",
              value=str(params.style_guide_additional_instructions),
              appearance="outline",
              style=me.Style(
                  margin=me.Margin.symmetric(horizontal=5), width="100%"
              ),
          )
        me.text(
            "Google Cloud Bucket URI containing supplementary materials"
            " (optional):",
            style=me.Style(margin=me.Margin(bottom=15)),
        )

        with components.row(
            align_items="center",
            gap=5,
            justify_content="space-between",
        ):
          with me.box(style=me.Style(width="40%")):
            me.input(
                label="GCP Bucket URI",
                on_blur=event_handlers.update_copycat_parameter,
                key="style_guide_files_uri",
                value=str(params.style_guide_files_uri),
                appearance="outline",
                style=me.Style(
                    margin=me.Margin.symmetric(horizontal=5), width="100%"
                ),
            )
          me.button(
              label="Generate",
              type="flat",
              on_click=event_handlers.generate_style_guide,
              style=me.Style(margin=me.Margin(top=20)),
          )

    with components.rounded_box_section(width="100%", height="80%"):
      with components.row(width="100%", height="100%", gap=0):
        # Markdown Editor Column
        with me.box(style=_STYLE_GUIDE_EDITOR_STYLE):
          me.native_textarea(
              value=params.style_guide,
              style=_STYLE_GUIDE_TEXTAREA_STYLE,
              on_blur=event_handlers.update_copycat_parameter,
              key="style_guide"
          )

        # Markdown Preview Column
        with me.box(style=_STYLE_GUIDE_PREVIEW_STYLE):
          me.markdown(params.style_guide)



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest
from absl.testing import parameterized
from vertexai import generative_models
from vertexai import language_models
import pandas as pd

from copycat import testing_utils


ad_copy_generator = testing_utils.ad_copy_generator
copycat = testing_utils.copycat


class PatchEmbeddingsModelTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.input_texts = [
        language_models.TextEmbeddingInput(
            text="Test 1", task_type="RETRIEVAL_DOCUMENT"
        ),
        language_models.TextEmbeddingInput(
            text="Test 2", task_type="RETRIEVAL_DOCUMENT"
        ),
    ]

  def test_random_embeddings_returns_expected_embeddings(self):
    # If this test fails then something has changed with the random number
    # generation used by the mock embedding model. This shouldn't happen,
    # but if it does just fix all tests, because it's nothing to do with
    # copycat.
    embeddings = testing_utils.random_embeddings(
        [
            language_models.TextEmbeddingInput(
                text="Test 1", task_type="RETRIEVAL_DOCUMENT"
            )
        ],
        output_dimensionality=5,
    )[0].values
    self.assertListEqual(
        embeddings,
        [
            0.2501313331234188,
            -0.4632126439646736,
            1.0048817101237066,
            0.53742307130886,
            -0.23405590135295984,
        ],
    )

  @testing_utils.PatchEmbeddingsModel()
  def test_get_embeddings_returns_list_of_embeddings(
      self, embeddings_model_patcher
  ):
    model = language_models.TextEmbeddingModel.from_pretrained("test_model")
    doc_embeddings = model.get_embeddings(
        self.input_texts, output_dimensionality=768
    )
    doc_embeddings = [emb.values for emb in doc_embeddings]

    self.assertLen(doc_embeddings, 2)
    self.assertLen(doc_embeddings[0], 768)
    self.assertLen(doc_embeddings[1], 768)

  @testing_utils.PatchEmbeddingsModel()
  def test_embeddings_are_deterministic(self, embeddings_model_patcher):
    model = language_models.TextEmbeddingModel.from_pretrained("test_model")
    doc_embeddings = model.get_embeddings(
        self.input_texts, output_dimensionality=768
    )
    doc_embeddings_2 = model.get_embeddings(
        self.input_texts, output_dimensionality=768
    )
    doc_embeddings = [emb.values for emb in doc_embeddings]
    doc_embeddings_2 = [emb.values for emb in doc_embeddings_2]
    self.assertSequenceEqual(doc_embeddings, doc_embeddings_2)

  @testing_utils.PatchEmbeddingsModel()
  def test_embeddings_depend_on_text_input(self, embeddings_model_patcher):
    model = language_models.TextEmbeddingModel.from_pretrained("test_model")
    doc_embeddings = model.get_embeddings(
        self.input_texts, output_dimensionality=768
    )
    doc_embeddings = [emb.values for emb in doc_embeddings]
    self.assertNotEqual(doc_embeddings[0], doc_embeddings[1])

  def test_patch_works_as_a_context_manager(self):
    with testing_utils.PatchEmbeddingsModel() as embeddings_model_patcher:
      model = language_models.TextEmbeddingModel.from_pretrained("test_model")
      doc_embeddings = model.get_embeddings(
          self.input_texts, output_dimensionality=768
      )
      doc_embeddings = [emb.values for emb in doc_embeddings]
      self.assertLen(doc_embeddings, 2)
      self.assertLen(doc_embeddings[0], 768)
      self.assertLen(doc_embeddings[1], 768)


class PatchGenerativeModelTest(parameterized.TestCase):

  def test_patch_works_as_a_context_manager(self):
    with testing_utils.PatchGenerativeModel(
        response="Response text"
    ) as generative_model_patcher:
      model = generative_models.GenerativeModel(model_name="test-model")
      response = model.generate_content("Example prompt")
      expected_response = generative_models.GenerationResponse.from_dict({
          "candidates": [{
              "finish_reason": generative_models.FinishReason.STOP,
              "content": {
                  "role": "model",
                  "parts": [{"text": "Response text"}],
              },
          }]
      })

      self.assertDictEqual(response.to_dict(), expected_response.to_dict())

      generative_model_patcher.mock_init.assert_called_once_with(
          model_name="test-model"
      )
      generative_model_patcher.mock_generative_model.generate_content.assert_called_once_with(
          "Example prompt"
      )

  @testing_utils.PatchGenerativeModel(response="Response text")
  def test_patch_works_as_a_decorator(self, generative_model_patcher):
    model = generative_models.GenerativeModel(model_name="test-model")
    response = model.generate_content("Example prompt")
    expected_response = generative_models.GenerationResponse.from_dict({
        "candidates": [{
            "finish_reason": generative_models.FinishReason.STOP,
            "content": {
                "role": "model",
                "parts": [{"text": "Response text"}],
            },
        }]
    })

    self.assertDictEqual(response.to_dict(), expected_response.to_dict())

    generative_model_patcher.mock_init.assert_called_once_with(
        model_name="test-model"
    )
    generative_model_patcher.mock_generative_model.generate_content.assert_called_once_with(
        "Example prompt"
    )

  def test_patch_can_produce_failed_response(self):
    with testing_utils.PatchGenerativeModel(
        response=generative_models.GenerationResponse.from_dict({
            "candidates": [{
                "finish_reason": generative_models.FinishReason.SAFETY,
            }]
        })
    ):
      expected_response = generative_models.GenerationResponse.from_dict({
          "candidates": [{
              "finish_reason": generative_models.FinishReason.SAFETY,
          }]
      })
      model = generative_models.GenerativeModel(model_name="test-model")
      self.assertDictEqual(
          model.generate_content("example prompt").to_dict(),
          expected_response.to_dict(),
      )


class ValuesAreEqualTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.embedding_model_patcher = testing_utils.PatchEmbeddingsModel()
    self.embedding_model_patcher.start()

  def tearDown(self):
    super().tearDown()
    self.embedding_model_patcher.stop()

  @parameterized.parameters(
      (1, 1, True),
      (1, 2, False),
      ([1], [1], True),
      ([1], [2], False),
      ([1], 1, False),
      ([1], (1,), True),
      ([[1, 2], 3], [[1, 2], 3], True),
      ([[1, 2], 3], [[3], 3], False),
      ({"a": 1}, {"a": 1}, True),
      ({"a": 1}, {"a": 2}, False),
      ({"a": 1}, {"b": 1}, False),
  )
  def test_values_are_equal(self, a, b, expected):
    self.assertEqual(testing_utils.values_are_equal(a, b), expected)

  @parameterized.parameters(
      (None, True),
      ("training_data", False),
      ("embedding_model_name", False),
      ("dimensionality", False),
      ("embeddings_batch_size", False),
  )
  def test_vectorstore_instances_are_equal(self, different_param, expected):
    params_1 = dict(
        training_data=pd.DataFrame({
            "headlines": [["headline 1"], ["headline 2"]],
            "descriptions": [["description 1"], ["description 2"]],
            "keywords": ["keyword 1", "keyword 2"],
        }),
        embedding_model_name="text-embedding-004",
        dimensionality=256,
        max_initial_ads=10,
        max_exemplar_ads=10,
        affinity_preference=-1,
        embeddings_batch_size=50,
    )

    params_2 = params_1.copy()
    if different_param:
      params_2[different_param] = dict(
          training_data=pd.DataFrame({
              "headlines": [["headline 3"], ["headline 2"]],
              "descriptions": [["description 1"], ["description 2"]],
              "keywords": ["keyword 1", "keyword 5"],
          }),
          embedding_model_name="text-multilingual-embedding-002",
          dimensionality=500,
          max_initial_ads=20,
          max_exemplar_ads=30,
          affinity_preference=-2,
          embeddings_batch_size=100,
      )[different_param]

    vectorstore_1 = ad_copy_generator.AdCopyVectorstore.create_from_pandas(
        **params_1
    )

    vectorstore_2 = ad_copy_generator.AdCopyVectorstore.create_from_pandas(
        **params_2
    )

    self.assertEqual(
        testing_utils.vectorstore_instances_are_equal(
            vectorstore_1, vectorstore_2
        ),
        expected,
    )

  def test_vectorstore_instances_are_equal_returns_false_for_different_types(
      self,
  ):
    vectorstore_1 = ad_copy_generator.AdCopyVectorstore.create_from_pandas(
        training_data=pd.DataFrame({
            "headlines": [["headline 1"], ["headline 2"]],
            "descriptions": [["description 1"], ["description 2"]],
            "keywords": ["keyword 1", "keyword 2"],
        }),
        embedding_model_name="text-embedding-004",
        dimensionality=256,
        max_initial_ads=10,
        max_exemplar_ads=10,
        affinity_preference=-1,
        embeddings_batch_size=50,
    )
    not_a_vectorstore = "not a vectorstore"
    self.assertFalse(
        testing_utils.vectorstore_instances_are_equal(
            vectorstore_1, not_a_vectorstore
        )
    )
    self.assertFalse(
        testing_utils.vectorstore_instances_are_equal(
            not_a_vectorstore, vectorstore_1
        )
    )

  @parameterized.parameters(
      (None, True),
      ("training_data", False),
      ("embedding_model_name", False),
      ("ad_format", False),
      ("embedding_model_dimensionality", False),
      ("embedding_model_batch_size", False),
  )
  def test_copycat_instances_are_equal(self, different_param, expected):
    params_1 = dict(
        training_data=pd.DataFrame({
            "headlines": [
                ["headline 1", "headline 2", "headline 3"],
                ["headline 2", "headline 3", "headline 4"],
            ],
            "descriptions": [
                ["description 1", "description 2"],
                ["description 2", "description 3"],
            ],
            "keywords": ["keyword 1", "keyword 2"],
        }),
        embedding_model_name="text-embedding-004",
        ad_format="text_ad",
    )

    params_2 = params_1.copy()
    if different_param:
      params_2[different_param] = dict(
          training_data=pd.DataFrame({
              "headlines": [
                  ["headline 1", "headline 4", "headline 3"],
                  ["headline 2", "headline 3", "headline 4"],
              ],
              "descriptions": [
                  ["description 1", "description 2"],
                  ["description 2", "description 5"],
              ],
              "keywords": ["keyword 1", "keyword 2"],
          }),
          embedding_model_name="text-multilingual-embedding-002",
          ad_format="responsive_search_ad",
          embedding_model_dimensionality=500,
          vectorstore_max_initial_ads=20,
          vectorstore_max_exemplar_ads=30,
          vectorstore_affinity_preference=-2,
          embedding_model_batch_size=100,
      )[different_param]

    copycat_1 = copycat.Copycat.create_from_pandas(**params_1)

    copycat_2 = copycat.Copycat.create_from_pandas(**params_2)

    self.assertEqual(
        testing_utils.copycat_instances_are_equal(copycat_1, copycat_2),
        expected,
    )

  def test_copycat_instances_are_equal_returns_false_for_different_types(
      self,
  ):
    copycat_1 = copycat.Copycat.create_from_pandas(
        training_data=pd.DataFrame({
            "headlines": [["headline 1"], ["headline 2"]],
            "descriptions": [["description 1"], ["description 2"]],
            "keywords": ["keyword 1", "keyword 2"],
        }),
        embedding_model_name="text-embedding-004",
        ad_format="text_ad",
        embedding_model_dimensionality=256,
        vectorstore_max_initial_ads=10,
        vectorstore_max_exemplar_ads=10,
        vectorstore_affinity_preference=-1,
        embedding_model_batch_size=50,
    )

    not_copycat = "not a copycat instance"

    self.assertFalse(
        testing_utils.copycat_instances_are_equal(copycat_1, not_copycat)
    )
    self.assertFalse(
        testing_utils.copycat_instances_are_equal(not_copycat, copycat_1)
    )


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import textwrap

from absl.testing import absltest
from absl.testing import parameterized
from vertexai import generative_models
import mock
import pandas as pd
import requests

from copycat import ad_copy_generator
from copycat import google_ads
from copycat import testing_utils


def mock_training_data(
    n_headlines_per_row: list[int], n_descriptions_per_row: list[int]
) -> pd.DataFrame:
  return pd.DataFrame({
      "headlines": [
          [f"train headline {i}_{j}" for i in range(n_headlines)]
          for j, n_headlines in enumerate(n_headlines_per_row)
      ],
      "descriptions": [
          [f"train description {i}_{j}" for i in range(n_descriptions)]
          for j, n_descriptions in enumerate(n_descriptions_per_row)
      ],
      "keywords": [
          f"keyword {i}a, keyword {i}b" for i in range(len(n_headlines_per_row))
      ],
  })


class TextGenerationRequestTest(parameterized.TestCase):

  def test_to_markdown_returns_expected_markdown(self):
    request = ad_copy_generator.TextGenerationRequest(
        keywords="keyword 1, keyword 2",
        prompt=[
            generative_models.Content(
                role="user",
                parts=[
                    generative_models.Part.from_text(
                        '{"Keywords": "keyword 1, keyword 2",'
                        ' "additional_instructions": ""}'
                    )
                ],
            ),
            generative_models.Content(
                role="model",
                parts=[
                    generative_models.Part.from_text(
                        '{"headlines":["headline 1","headline 2"],'
                        '"descriptions":["description 1","description 2"]}'
                    )
                ],
            ),
            generative_models.Content(
                role="user",
                parts=[
                    generative_models.Part.from_text(
                        '{"Keywords": "keyword 3, keyword 4",'
                        ' "additional_instructions": "something"}'
                    )
                ],
            ),
        ],
        system_instruction="My system instruction",
        chat_model_name=ad_copy_generator.ModelName.GEMINI_1_5_FLASH,
        temperature=0.95,
        top_k=20,
        top_p=0.95,
        safety_settings=None,
        existing_ad_copy=google_ads.GoogleAd(headlines=[], descriptions=[]),
    )

    expected_markdown = textwrap.dedent(
        """\
      **Keywords:**
      
      keyword 1, keyword 2
      
      **Model Parameters:**

      Model name: gemini-1.5-flash-002

      Temperature: 0.95

      Top K: 20

      Top P: 0.95
      
      Safety settings: None

      **System instruction:**

      My system instruction

      **User:**

      {"Keywords": "keyword 1, keyword 2", "additional_instructions": ""}

      **Model:**

      {"headlines":["headline 1","headline 2"],"descriptions":["description 1","description 2"]}

      **User:**

      {"Keywords": "keyword 3, keyword 4", "additional_instructions": "something"}"""
    )
    self.assertEqual(request.to_markdown(), expected_markdown)

  def test_to_markdown_returns_expected_markdown_with_existing_ad_copy(self):
    request = ad_copy_generator.TextGenerationRequest(
        keywords="keyword 1, keyword 2",
        prompt=[
            generative_models.Content(
                role="user",
                parts=[
                    generative_models.Part.from_text(
                        '{"Keywords": "keyword 1, keyword 2",'
                        ' "additional_instructions": ""}'
                    )
                ],
            ),
            generative_models.Content(
                role="model",
                parts=[
                    generative_models.Part.from_text(
                        '{"headlines":["headline 1","headline 2"],'
                        '"descriptions":["description 1","description 2"]}'
                    )
                ],
            ),
            generative_models.Content(
                role="user",
                parts=[
                    generative_models.Part.from_text(
                        '{"Keywords": "keyword 3, keyword 4",'
                        ' "additional_instructions": "something"}'
                    )
                ],
            ),
        ],
        system_instruction="My system instruction",
        chat_model_name=ad_copy_generator.ModelName.GEMINI_1_5_FLASH,
        temperature=0.95,
        top_k=20,
        top_p=0.95,
        safety_settings=None,
        existing_ad_copy=google_ads.GoogleAd(
            headlines=["headline 1"], descriptions=["description 1"]
        ),
    )

    expected_markdown = textwrap.dedent(
        """\
      **Keywords:**

      keyword 1, keyword 2

      **Existing headlines:**

      ['headline 1']

      **Existing descriptions:**

      ['description 1']

      **Model Parameters:**

      Model name: gemini-1.5-flash-002

      Temperature: 0.95

      Top K: 20

      Top P: 0.95

      Safety settings: None

      **System instruction:**

      My system instruction

      **User:**

      {"Keywords": "keyword 1, keyword 2", "additional_instructions": ""}

      **Model:**

      {"headlines":["headline 1","headline 2"],"descriptions":["description 1","description 2"]}

      **User:**

      {"Keywords": "keyword 3, keyword 4", "additional_instructions": "something"}"""
    )
    self.assertEqual(request.to_markdown(), expected_markdown)


class AdCopyVectorstoreTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.tmp_dir = self.create_tempdir()
    self.embedding_model_patcher = testing_utils.PatchEmbeddingsModel()
    self.embedding_model_patcher.start()

  def tearDown(self):
    super().tearDown()
    self.embedding_model_patcher.stop()

  def test_create_from_pandas_deduplicates_ads(self):
    # Training data has same ad for two different sets of keywords.
    # It should keep only one of them.
    training_data = pd.DataFrame.from_records([
        {
            "headlines": ["headline 1", "headline 2"],
            "descriptions": ["description 1", "description 2"],
            "keywords": "keyword 1, keyword 2",
        },
        {
            "headlines": ["headline 1", "headline 2"],
            "descriptions": ["description 1", "description 2"],
            "keywords": "keyword 3, keyword 4",
        },
    ])

    ad_copy_vectorstore = (
        ad_copy_generator.AdCopyVectorstore.create_from_pandas(
            training_data=training_data,
            embedding_model_name="text-embedding-004",
            dimensionality=256,
            max_initial_ads=100,
            max_exemplar_ads=10,
            affinity_preference=None,
            embeddings_batch_size=10,
            exemplar_selection_method="random",
        )
    )
    self.assertLen(
        ad_copy_vectorstore.ad_exemplars,
        1,
    )

  def test_get_relevant_ads_retrieves_keywords_and_ads(self):

    training_data = pd.DataFrame.from_records([
        {
            "headlines": ["headline 1", "headline 2"],
            "descriptions": ["description 1", "description 2"],
            "keywords": "keyword 1, keyword 2",
        },
        {
            "headlines": ["headline 3"],
            "descriptions": ["description 3"],
            "keywords": "keyword 3, keyword 4",
        },
        {
            "headlines": ["headline 4", "headline 5"],
            "descriptions": ["description 2"],
            "keywords": "keyword 5, keyword 6",
        },
    ])

    ad_copy_vectorstore = (
        ad_copy_generator.AdCopyVectorstore.create_from_pandas(
            training_data=training_data,
            embedding_model_name="text-embedding-004",
            dimensionality=256,
            max_initial_ads=100,
            max_exemplar_ads=10,
            affinity_preference=None,
            embeddings_batch_size=10,
            exemplar_selection_method="random",
        )
    )

    results = ad_copy_vectorstore.get_relevant_ads(["test query"], k=2)
    expected_results = [[
        ad_copy_generator.ExampleAd(
            keywords="keyword 1, keyword 2",
            google_ad=google_ads.GoogleAd(
                headlines=["headline 1", "headline 2"],
                descriptions=["description 1", "description 2"],
            ),
        ),
        ad_copy_generator.ExampleAd(
            keywords="keyword 3, keyword 4",
            google_ad=google_ads.GoogleAd(
                headlines=["headline 3"],
                descriptions=["description 3"],
            ),
        ),
    ]]

    self.assertListEqual(results, expected_results)

  def test_get_relevant_ads_and_embeddings_from_embeddings(self):
    ad_exemplars = pd.DataFrame.from_records([
        {
            "headlines": ["headline 1", "headline 2"],
            "descriptions": ["description 1", "description 2"],
            "keywords": "keyword 1, keyword 2",
            "embeddings": [1.0, 2.0, 3.0],
        },
        {
            "headlines": ["headline 3"],
            "descriptions": ["description 3"],
            "keywords": "keyword 3, keyword 4",
            "embeddings": [4.0, 5.0, 6.0],
        },
        {
            "headlines": ["headline 4", "headline 5"],
            "descriptions": ["description 2"],
            "keywords": "keyword 5, keyword 6",
            "embeddings": [7.0, 8.0, 9.0],
        },
    ])

    ad_copy_vectorstore = ad_copy_generator.AdCopyVectorstore(
        ad_exemplars=ad_exemplars,
        embedding_model_name="text-embedding-004",
        dimensionality=3,
        embeddings_batch_size=10,
    )

    results = (
        ad_copy_vectorstore.get_relevant_ads_and_embeddings_from_embeddings(
            [[2.0, 3.0, 4.0]], k=2
        )
    )
    expected_ads = [[
        ad_copy_generator.ExampleAd(
            keywords="keyword 1, keyword 2",
            google_ad=google_ads.GoogleAd(
                headlines=["headline 1", "headline 2"],
                descriptions=["description 1", "description 2"],
            ),
        ),
        ad_copy_generator.ExampleAd(
            keywords="keyword 3, keyword 4",
            google_ad=google_ads.GoogleAd(
                headlines=["headline 3"],
                descriptions=["description 3"],
            ),
        ),
    ]]
    expected_embeddings = [[[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]]]
    expected_results = (expected_ads, expected_embeddings)

    self.assertEqual(results, expected_results)

  def test_affinity_propagation_is_used_to_select_ads_if_provided(self):
    training_data = pd.DataFrame.from_records([
        {
            "headlines": ["headline 1", "headline 2"],
            "descriptions": ["description 1", "description 2"],
            "keywords": "keyword 1, keyword 2",
        },
        {
            "headlines": ["headline 3"],
            "descriptions": ["description 3"],
            "keywords": "keyword 3, keyword 4",
        },
    ])

    with mock.patch(
        "google3.third_party.professional_services.solutions.copycat.ad_copy_generator.cluster.AffinityPropagation"
    ) as mock_affinity_propagation:
      (
          ad_copy_generator.AdCopyVectorstore.create_from_pandas(
              training_data=training_data,
              embedding_model_name="text-embedding-004",
              dimensionality=256,
              max_initial_ads=100,
              max_exemplar_ads=10,
              affinity_preference=None,
              embeddings_batch_size=10,
              exemplar_selection_method="affinity_propagation",
          )
      )

      mock_affinity_propagation.assert_called_once()

  def test_to_dict_and_from_dict_returns_same_ad_copy_vectorstore(self):
    training_data = pd.DataFrame.from_records([
        {
            "headlines": ["headline 1", "headline 2"],
            "descriptions": ["description 1", "description 2"],
            "keywords": "keyword 1, keyword 2",
        },
        {
            "headlines": ["headline 3"],
            "descriptions": ["description 3"],
            "keywords": "keyword 3, keyword 4",
        },
        {
            "headlines": ["headline 4", "headline 5"],
            "descriptions": ["description 2"],
            "keywords": "keyword 5, keyword 6",
        },
    ])

    ad_copy_vectorstore = (
        ad_copy_generator.AdCopyVectorstore.create_from_pandas(
            training_data=training_data,
            embedding_model_name="text-embedding-004",
            dimensionality=256,
            max_initial_ads=100,
            max_exemplar_ads=10,
            affinity_preference=None,
            embeddings_batch_size=10,
            exemplar_selection_method="random",
        )
    )

    reloaded_ad_copy_vectorstore = (
        ad_copy_generator.AdCopyVectorstore.from_dict(
            ad_copy_vectorstore.to_dict()
        )
    )

    self.assertTrue(
        testing_utils.vectorstore_instances_are_equal(
            ad_copy_vectorstore, reloaded_ad_copy_vectorstore
        )
    )

  @parameterized.parameters([
      "embedding_model_name",
      "dimensionality",
      "embeddings_batch_size",
      "ad_exemplars",
  ])
  def test_from_dict_raises_key_error_if_required_key_is_missing(
      self, required_key
  ):
    with self.assertRaises(KeyError):
      training_data = pd.DataFrame.from_records([
          {
              "headlines": ["headline 1", "headline 2"],
              "descriptions": ["description 1", "description 2"],
              "keywords": "keyword 1, keyword 2",
          },
          {
              "headlines": ["headline 3"],
              "descriptions": ["description 3"],
              "keywords": "keyword 3, keyword 4",
          },
          {
              "headlines": ["headline 4", "headline 5"],
              "descriptions": ["description 2"],
              "keywords": "keyword 5, keyword 6",
          },
      ])

      ad_copy_vectorstore = (
          ad_copy_generator.AdCopyVectorstore.create_from_pandas(
              training_data=training_data,
              embedding_model_name="text-embedding-004",
              dimensionality=256,
              max_initial_ads=100,
              max_exemplar_ads=10,
              affinity_preference=None,
              embeddings_batch_size=10,
              exemplar_selection_method="random",
          )
      )

      ad_copy_vectorstore_dict = ad_copy_vectorstore.to_dict()
      del ad_copy_vectorstore_dict[required_key]
      ad_copy_generator.AdCopyVectorstore.from_dict(ad_copy_vectorstore_dict)

  def test_to_json_and_from_json_returns_same_ad_copy_vectorstore(self):
    training_data = pd.DataFrame.from_records([
        {
            "headlines": ["headline 1", "headline 2"],
            "descriptions": ["description 1", "description 2"],
            "keywords": "keyword 1, keyword 2",
        },
        {
            "headlines": ["headline 3"],
            "descriptions": ["description 3"],
            "keywords": "keyword 3, keyword 4",
        },
        {
            "headlines": ["headline 4", "headline 5"],
            "descriptions": ["description 2"],
            "keywords": "keyword 5, keyword 6",
        },
    ])

    ad_copy_vectorstore = (
        ad_copy_generator.AdCopyVectorstore.create_from_pandas(
            training_data=training_data,
            embedding_model_name="text-embedding-004",
            dimensionality=256,
            max_initial_ads=100,
            max_exemplar_ads=10,
            affinity_preference=None,
            embeddings_batch_size=10,
            exemplar_selection_method="random",
        )
    )

    reloaded_ad_copy_vectorstore = (
        ad_copy_generator.AdCopyVectorstore.from_json(
            ad_copy_vectorstore.to_json()
        )
    )

    self.assertTrue(
        testing_utils.vectorstore_instances_are_equal(
            ad_copy_vectorstore, reloaded_ad_copy_vectorstore
        )
    )

  def test_unique_headlines_and_descriptions_are_set_correctly(self):
    training_data = pd.DataFrame.from_records([
        {
            "headlines": ["headline 1", "headline 2"],
            "descriptions": ["description 1", "description 2"],
            "keywords": "keyword 1, keyword 2",
        },
        {
            "headlines": ["headline 3", "headline 2"],
            "descriptions": ["description 3", "description 1"],
            "keywords": "keyword 3, keyword 4",
        },
    ])
    ad_copy_vectorstore = (
        ad_copy_generator.AdCopyVectorstore.create_from_pandas(
            training_data=training_data,
            embedding_model_name="text-embedding-004",
            dimensionality=256,
            max_initial_ads=100,
            max_exemplar_ads=10,
            affinity_preference=None,
            embeddings_batch_size=10,
            exemplar_selection_method="random",
        )
    )

    self.assertSetEqual(
        ad_copy_vectorstore.unique_headlines,
        {"headline 1", "headline 2", "headline 3"},
    )
    self.assertSetEqual(
        ad_copy_vectorstore.unique_descriptions,
        {"description 1", "description 2", "description 3"},
    )

  def test_embed_documents_returns_expected_embeddings(self):
    training_data = pd.DataFrame.from_records([
        {
            "headlines": ["headline 1", "headline 2"],
            "descriptions": ["description 1", "description 2"],
            "keywords": "keyword 1, keyword 2",
        },
        {
            "headlines": ["headline 3"],
            "descriptions": ["description 3"],
            "keywords": "keyword 3, keyword 4",
        },
    ])
    ad_copy_vectorstore = (
        ad_copy_generator.AdCopyVectorstore.create_from_pandas(
            training_data=training_data,
            embedding_model_name="text-embedding-004",
            dimensionality=256,
            max_initial_ads=100,
            max_exemplar_ads=10,
            affinity_preference=None,
            embeddings_batch_size=10,
            exemplar_selection_method="random",
        )
    )

    embeddings = ad_copy_vectorstore.embed_documents(
        ["headline 1", "headline 2", "headline 3"]
    )
    self.assertLen(embeddings, 3)
    self.assertLen(embeddings[0], 256)
    self.assertLen(embeddings[1], 256)
    self.assertLen(embeddings[2], 256)

  def test_embed_queries_embeds_queries_correctly(self):
    training_data = pd.DataFrame.from_records([
        {
            "headlines": ["headline 1", "headline 2"],
            "descriptions": ["description 1", "description 2"],
            "keywords": "keyword 1, keyword 2",
        },
        {
            "headlines": ["headline 3"],
            "descriptions": ["description 3"],
            "keywords": "keyword 3, keyword 4",
        },
    ])
    ad_copy_vectorstore = (
        ad_copy_generator.AdCopyVectorstore.create_from_pandas(
            training_data=training_data,
            embedding_model_name="text-embedding-004",
            dimensionality=256,
            max_initial_ads=100,
            max_exemplar_ads=10,
            affinity_preference=None,
            embeddings_batch_size=10,
            exemplar_selection_method="random",
        )
    )

    embeddings = ad_copy_vectorstore.embed_queries(
        ["headline 1", "headline 2", "headline 3"]
    )
    self.assertLen(embeddings, 3)
    self.assertLen(embeddings[0], 256)
    self.assertLen(embeddings[1], 256)
    self.assertLen(embeddings[2], 256)


class AdCopyGeneratorTest(parameterized.TestCase):

  def test_construct_system_instruction_constructs_expected_instruction_with_style_guide(
      self,
  ):

    instruction = ad_copy_generator.construct_system_instruction(
        system_instruction="My system instruction",
        style_guide="My style guide",
        system_instruction_kwargs={},
    )

    expected_instruction = "My system instruction\n\nMy style guide"
    self.assertEqual(instruction, expected_instruction)

  def test_construct_system_instruction_replaces_system_instruction_placeholders(
      self,
  ):
    instruction = ad_copy_generator.construct_system_instruction(
        system_instruction="My system instruction for {company_name}",
        style_guide="",
        system_instruction_kwargs=dict(
            company_name="my company",
        ),
    )

    expected_instruction = "My system instruction for my company"
    self.assertEqual(instruction, expected_instruction)

  def test_construct_system_instruction_constructs_expected_instruction_without_style_guide(
      self,
  ):

    instruction = ad_copy_generator.construct_system_instruction(
        system_instruction="My system instruction",
        style_guide="",
        system_instruction_kwargs={},
    )
    expected_instruction = "My system instruction"
    self.assertEqual(instruction, expected_instruction)

  @parameterized.named_parameters([
      {
          "testcase_name": (
              "without keyword specific instructions or existing ad copy"
          ),
          "existing_ad_copy": None,
          "keyword_specific_instructions": "",
          "expected_final_message": (
              "Please write 3 headlines and 2 descriptions for this"
              " ad.\n\nKeywords: Keyword 1, Keyword 2, keyword λ"
          ),
      },
      {
          "testcase_name": "with keyword specific instructions",
          "existing_ad_copy": None,
          "keyword_specific_instructions": (
              "My keywords specific instructions with unicode ß"
          ),
          "expected_final_message": (
              "For the next set of keywords, please consider the"
              " following additional instructions:\n\nMy keywords"
              " specific instructions with unicode ß\n\nPlease write 3"
              " headlines and 2 descriptions for this ad.\n\nKeywords:"
              " Keyword 1, Keyword 2, keyword λ"
          ),
      },
      {
          "testcase_name": "with existing headlines",
          "existing_ad_copy": google_ads.GoogleAd(
              headlines=["existing headline"], descriptions=[]
          ),
          "keyword_specific_instructions": "",
          "expected_final_message": (
              "This ad already has 1 headlines.\n\n- headlines: ['existing"
              " headline']\n\nPlease write 2 more headlines and 2 more"
              " descriptions to complete this ad.\n\nKeywords: Keyword 1,"
              " Keyword 2, keyword λ"
          ),
      },
      {
          "testcase_name": "with existing descriptions",
          "existing_ad_copy": google_ads.GoogleAd(
              headlines=[], descriptions=["existing description"]
          ),
          "keyword_specific_instructions": "",
          "expected_final_message": (
              "This ad already has 1 descriptions.\n\n- descriptions:"
              " ['existing description']\n\nPlease write 3 more headlines and 1"
              " more descriptions to complete this ad.\n\nKeywords: Keyword 1,"
              " Keyword 2, keyword λ"
          ),
      },
      {
          "testcase_name": "with existing headlines and descriptions",
          "existing_ad_copy": google_ads.GoogleAd(
              headlines=["existing headline"],
              descriptions=["existing description"],
          ),
          "keyword_specific_instructions": "",
          "expected_final_message": (
              "This ad already has 1 headlines and 1 descriptions:\n\n-"
              " headlines: ['existing headline']\n- descriptions: ['existing"
              " description']\n\nPlease write 2 more headlines and 1 more"
              " descriptions to complete this ad.\n\nKeywords: Keyword 1,"
              " Keyword 2, keyword λ"
          ),
      },
      {
          "testcase_name": "with complete headlines",
          "existing_ad_copy": google_ads.GoogleAd(
              headlines=[
                  "existing headline 1",
                  "existing headline 2",
                  "existing headline 3",
              ],
              descriptions=[],
          ),
          "keyword_specific_instructions": "",
          "expected_final_message": (
              "This ad already has 3 headlines.\n\n- headlines: ['existing"
              " headline 1', 'existing headline 2', 'existing headline"
              " 3']\n\nPlease write 2 more descriptions to complete this ad."
              " You do not need to write any headlines, as there are enough"
              " already.\n\nKeywords: Keyword 1, Keyword 2, keyword λ"
          ),
      },
      {
          "testcase_name": "with complete descriptions",
          "existing_ad_copy": google_ads.GoogleAd(
              headlines=[],
              descriptions=["existing description 1", "existing description 2"],
          ),
          "keyword_specific_instructions": "",
          "expected_final_message": (
              "This ad already has 2 descriptions.\n\n- descriptions:"
              " ['existing description 1', 'existing description 2']\n\nPlease"
              " write 3 more headlines to complete this ad. You do not need to"
              " write any descriptions, as there are enough"
              " already.\n\nKeywords: Keyword 1, Keyword 2, keyword λ"
          ),
      },
      {
          "testcase_name": "with complete descriptions and some headlines",
          "existing_ad_copy": google_ads.GoogleAd(
              headlines=["existing headline"],
              descriptions=["existing description 1", "existing description 2"],
          ),
          "keyword_specific_instructions": "",
          "expected_final_message": (
              "This ad already has 1 headlines and 2 descriptions:\n\n-"
              " headlines: ['existing headline']\n- descriptions: ['existing"
              " description 1', 'existing description 2']\n\nPlease write 2"
              " more headlines to complete this ad. You do not need to write"
              " any descriptions, as there are enough already.\n\nKeywords:"
              " Keyword 1, Keyword 2, keyword λ"
          ),
      },
      {
          "testcase_name": "with complete headlines and some descriptions",
          "existing_ad_copy": google_ads.GoogleAd(
              headlines=[
                  "existing headline 1",
                  "existing headline 2",
                  "existing headline 3",
              ],
              descriptions=["existing description"],
          ),
          "keyword_specific_instructions": "",
          "expected_final_message": (
              "This ad already has 3 headlines and 1 descriptions:\n\n-"
              " headlines: ['existing headline 1', 'existing headline 2',"
              " 'existing headline 3']\n- descriptions: ['existing"
              " description']\n\nPlease write 1 more descriptions to complete"
              " this ad. You do not need to write any headlines, as there are"
              " enough already.\n\nKeywords: Keyword 1, Keyword 2, keyword λ"
          ),
      },
  ])
  def test_construct_new_ad_copy_prompt_constructs_expected_prompt(
      self,
      existing_ad_copy,
      keyword_specific_instructions,
      expected_final_message,
  ):
    prompt = ad_copy_generator.construct_new_ad_copy_prompt(
        example_ads=[
            ad_copy_generator.ExampleAd(
                keywords="keyword 5, keyword 6",
                google_ad=google_ads.GoogleAd(
                    headlines=["headline 4", "headline 5"],
                    descriptions=["description 2"],
                ),
            ),
            ad_copy_generator.ExampleAd(
                keywords="keyword 3, keyword 4",
                google_ad=google_ads.GoogleAd(
                    headlines=["headline 3"],
                    descriptions=[
                        "description 3",
                        "description with unicode weiß",
                    ],
                ),
            ),
        ],
        keywords_specific_instructions=keyword_specific_instructions,
        keywords="Keyword 1, Keyword 2, keyword λ",
        ad_format=google_ads.TEXT_AD_FORMAT,
        existing_ad_copy=existing_ad_copy,
    )

    expected_prompt = [
        generative_models.Content(
            role="user",
            parts=[
                generative_models.Part.from_text(
                    "Please write 1 headlines and 2 descriptions for this"
                    " ad.\n\nKeywords: keyword 3, keyword 4"
                )
            ],
        ),
        generative_models.Content(
            role="model",
            parts=[
                generative_models.Part.from_text(
                    '{"headlines":["headline 3"],"descriptions":'
                    '["description 3","description with unicode weiß"]}'
                )
            ],
        ),
        generative_models.Content(
            role="user",
            parts=[
                generative_models.Part.from_text(
                    "Please write 2 headlines and 1 descriptions for this"
                    " ad.\n\nKeywords: keyword 5, keyword 6"
                )
            ],
        ),
        generative_models.Content(
            role="model",
            parts=[
                generative_models.Part.from_text(
                    '{"headlines":["headline 4","headline 5"],'
                    '"descriptions":["description 2"]}'
                )
            ],
        ),
        generative_models.Content(
            role="user",
            parts=[generative_models.Part.from_text(expected_final_message)],
        ),
    ]

    for single_prompt, single_expected_prompt in zip(prompt, expected_prompt):
      self.assertEqual(
          single_prompt.to_dict(), single_expected_prompt.to_dict()
      )

  def test_construct_new_ad_copy_prompt_raises_value_error_for_complete_ad(
      self,
  ):
    with self.assertRaisesWithLiteralMatch(
        ValueError, "Trying to generate an ad that is already complete."
    ):
      ad_copy_generator.construct_new_ad_copy_prompt(
          example_ads=[
              ad_copy_generator.ExampleAd(
                  keywords="keyword 5, keyword 6",
                  google_ad=google_ads.GoogleAd(
                      headlines=["headline 4", "headline 5"],
                      descriptions=["description 2"],
                  ),
              ),
              ad_copy_generator.ExampleAd(
                  keywords="keyword 3, keyword 4",
                  google_ad=google_ads.GoogleAd(
                      headlines=["headline 3"],
                      descriptions=[
                          "description 3",
                          "description with unicode weiß",
                      ],
                  ),
              ),
          ],
          keywords_specific_instructions="",
          keywords="Keyword 1, Keyword 2, keyword λ",
          ad_format=google_ads.TEXT_AD_FORMAT,
          existing_ad_copy=google_ads.GoogleAd(
              headlines=["headline 1", "headline 2", "headline 3"],
              descriptions=["description 1", "description 2"],
          ),
      )

  @parameterized.named_parameters([
      {
          "testcase_name": "too many headlines",
          "headlines": [f"headline {i}" for i in range(16)],
          "descriptions": ["description 1"],
          "fixed_headlines": [f"headline {i}" for i in range(15)],
          "fixed_descriptions": ["description 1"],
      },
      {
          "testcase_name": "too many descriptions",
          "headlines": ["headline 1"],
          "descriptions": [f"description {i}" for i in range(6)],
          "fixed_headlines": ["headline 1"],
          "fixed_descriptions": [f"description {i}" for i in range(4)],
      },
      {
          "testcase_name": "too long headline",
          "headlines": ["headline 1", "a" * 31, "headline 2"],
          "descriptions": ["description 1"],
          "fixed_headlines": ["headline 1", "headline 2"],
          "fixed_descriptions": ["description 1"],
      },
      {
          "testcase_name": "too long description",
          "headlines": ["headline 1", "headline 2"],
          "descriptions": ["description 1", "a" * 91, "description 2"],
          "fixed_headlines": ["headline 1", "headline 2"],
          "fixed_descriptions": ["description 1", "description 2"],
      },
      {
          "testcase_name": "dynamic keyword insertion",
          "headlines": [
              "Valid DKI {KeyWord:my keyword} 123",
              "Invalid DKI {KeyWord:my keyword} too long!",
          ],
          "descriptions": [
              "Valid DKI {KeyWord:my keyword} " + "a" * 63,
              "Invalid DKI {KeyWord:my keyword} too long!" + "a" * 63,
          ],
          "fixed_headlines": ["Valid DKI {KeyWord:my keyword} 123"],
          "fixed_descriptions": ["Valid DKI {KeyWord:my keyword} " + "a" * 63],
      },
  ])
  def test_remove_invalid_headlines_and_descriptions_returns_fixed_ad_copy(
      self, headlines, descriptions, fixed_headlines, fixed_descriptions
  ):
    google_ad = google_ads.GoogleAd(
        headlines=headlines,
        descriptions=descriptions,
    )
    ad_copy_generator.remove_invalid_headlines_and_descriptions(
        google_ad, google_ads.RESPONSIVE_SEARCH_AD_FORMAT
    )

    expected_fixed_ad = google_ads.GoogleAd(
        headlines=fixed_headlines,
        descriptions=fixed_descriptions,
    )
    self.assertEqual(expected_fixed_ad, google_ad)

  @parameterized.parameters(
      (
          ad_copy_generator.ModelName.GEMINI_1_5_FLASH,
          "gemini-1.5-flash-002",
      ),
      (
          ad_copy_generator.ModelName.GEMINI_1_5_PRO,
          "gemini-1.5-pro-002",
      ),
  )
  @testing_utils.PatchGenerativeModel(response="Response text")
  def test_generate_google_ad_parses_model_name_correctly(
      self, input_model_name, parsed_model_name, generative_model_patcher
  ):
    request = ad_copy_generator.TextGenerationRequest(
        keywords="keyword 1, keyword 2",
        prompt=[
            generative_models.Content(
                role="user",
                parts=[generative_models.Part.from_text("Example prompt")],
            )
        ],
        system_instruction="Example system instruction",
        chat_model_name=input_model_name,
        temperature=0.9,
        top_k=40,
        top_p=0.95,
        safety_settings=None,
        existing_ad_copy=google_ads.GoogleAd(headlines=[], descriptions=[]),
    )
    ad_copy_generator.generate_google_ad_json_batch([request])

    self.assertEqual(
        generative_model_patcher.mock_init.call_args[1]["model_name"],
        parsed_model_name,
    )

  @testing_utils.PatchGenerativeModel(response="Response text")
  def test_generate_google_ad_parses_safety_settings_correctly(
      self, generative_model_patcher
  ):
    custom_safety_settings = {
        ad_copy_generator.generative_models.HarmCategory.HARM_CATEGORY_HATE_SPEECH: (
            ad_copy_generator.generative_models.HarmBlockThreshold.BLOCK_NONE
        )
    }
    request = ad_copy_generator.TextGenerationRequest(
        keywords="keyword 1, keyword 2",
        prompt=[
            generative_models.Content(
                role="user",
                parts=[generative_models.Part.from_text("Example prompt")],
            )
        ],
        system_instruction="Example system instruction",
        chat_model_name=ad_copy_generator.ModelName.GEMINI_1_5_FLASH,
        temperature=0.9,
        top_k=40,
        top_p=0.95,
        safety_settings=custom_safety_settings,
        existing_ad_copy=google_ads.GoogleAd(headlines=[], descriptions=[]),
    )
    ad_copy_generator.generate_google_ad_json_batch([request])

    self.assertEqual(
        generative_model_patcher.mock_init.call_args[1]["safety_settings"],
        custom_safety_settings,
    )

  @parameterized.parameters(
      ad_copy_generator.ModelName.GEMINI_1_5_FLASH,
      ad_copy_generator.ModelName.GEMINI_1_5_PRO,
  )
  @testing_utils.PatchGenerativeModel(response="Response text")
  def test_generate_google_ad_uses_expected_generation_config(
      self, input_model_name, generative_model_patcher
  ):

    request = ad_copy_generator.TextGenerationRequest(
        keywords="keyword 1, keyword 2",
        prompt=[
            generative_models.Content(
                role="user",
                parts=[generative_models.Part.from_text("Example prompt")],
            )
        ],
        system_instruction="Example system instruction",
        chat_model_name=input_model_name,
        temperature=0.9,
        top_k=40,
        top_p=0.95,
        safety_settings=None,
        existing_ad_copy=google_ads.GoogleAd(headlines=[], descriptions=[]),
    )
    ad_copy_generator.generate_google_ad_json_batch([request])

    expected_generation_config = dict(
        temperature=0.9,
        top_k=40,
        top_p=0.95,
        response_mime_type="application/json",
    )
    expected_generation_config["response_schema"] = {
        "properties": {
            "headlines": {
                "items": {
                    "description": (
                        "The headlines for the ad. Must be fewer than 30"
                        " characters."
                    ),
                    "type": "STRING",
                },
                "type": "ARRAY",
            },
            "descriptions": {
                "items": {
                    "description": (
                        "The descriptions for the ad. Must be fewer than 90"
                        " characters."
                    ),
                    "type": "STRING",
                },
                "type": "ARRAY",
            },
        },
        "required": ["headlines", "descriptions"],
        "property_ordering": ["headlines", "descriptions"],
        "type": "OBJECT",
    }

    self.assertDictEqual(
        generative_model_patcher.mock_init.call_args[1][
            "generation_config"
        ].to_dict(),
        expected_generation_config,
    )

  @parameterized.parameters(
      ad_copy_generator.ModelName.GEMINI_1_5_FLASH,
      ad_copy_generator.ModelName.GEMINI_1_5_PRO,
  )
  @testing_utils.PatchGenerativeModel(response="Response text")
  def test_generate_google_ad_uses_prompt(
      self, input_model_name, generative_model_patcher
  ):
    prompt = [
        generative_models.Content(
            role="user",
            parts=[generative_models.Part.from_text("Example prompt")],
        )
    ]
    request = ad_copy_generator.TextGenerationRequest(
        keywords="keyword 1, keyword 2",
        prompt=prompt,
        system_instruction="Example system instruction",
        chat_model_name=input_model_name,
        temperature=0.9,
        top_k=40,
        top_p=0.95,
        safety_settings=None,
        existing_ad_copy=google_ads.GoogleAd(headlines=[], descriptions=[]),
    )
    ad_copy_generator.generate_google_ad_json_batch([request])

    generative_model_patcher.mock_generative_model.generate_content_async.assert_called_with(
        prompt
    )

  @parameterized.parameters(
      ad_copy_generator.ModelName.GEMINI_1_5_FLASH,
      ad_copy_generator.ModelName.GEMINI_1_5_PRO,
  )
  @testing_utils.PatchGenerativeModel(response="Response text")
  def test_generate_google_ad_returns_response_from_gemini(
      self, input_model_name, generative_model_patcher
  ):

    request = ad_copy_generator.TextGenerationRequest(
        keywords="keyword 1, keyword 2",
        prompt=[
            generative_models.Content(
                role="user",
                parts=[generative_models.Part.from_text("Example prompt")],
            )
        ],
        system_instruction="Example system instruction",
        chat_model_name=input_model_name,
        temperature=0.9,
        top_k=40,
        top_p=0.95,
        safety_settings=None,
        existing_ad_copy=google_ads.GoogleAd(headlines=[], descriptions=[]),
    )
    response = ad_copy_generator.generate_google_ad_json_batch([request])
    self.assertEqual(
        response[0].candidates[0].content.parts[0].text, "Response text"
    )

  @parameterized.parameters(
      ad_copy_generator.ModelName.GEMINI_1_5_FLASH,
      ad_copy_generator.ModelName.GEMINI_1_5_PRO,
  )
  @testing_utils.PatchGenerativeModel(response="Response text")
  def test_generate_google_ad_returns_multiple_responses_from_gemini_in_batch(
      self, input_model_name, generative_model_patcher
  ):

    prompt_1 = [
        generative_models.Content(
            role="user",
            parts=[generative_models.Part.from_text("Example prompt")],
        )
    ]
    prompt_2 = [
        generative_models.Content(
            role="user",
            parts=[generative_models.Part.from_text("Another example prompt")],
        )
    ]

    request_1 = ad_copy_generator.TextGenerationRequest(
        keywords="keyword 1, keyword 2",
        prompt=prompt_1,
        system_instruction="Example system instruction",
        chat_model_name=input_model_name,
        temperature=0.9,
        top_k=40,
        top_p=0.95,
        safety_settings=None,
        existing_ad_copy=google_ads.GoogleAd(headlines=[], descriptions=[]),
    )

    request_2 = ad_copy_generator.TextGenerationRequest(
        keywords="keyword 1, keyword 2",
        prompt=prompt_2,
        system_instruction="Example system instruction",
        chat_model_name=input_model_name,
        temperature=0.9,
        top_k=40,
        top_p=0.95,
        safety_settings=None,
        existing_ad_copy=google_ads.GoogleAd(headlines=[], descriptions=[]),
    )

    response = ad_copy_generator.generate_google_ad_json_batch(
        [request_1, request_2]
    )

    self.assertLen(response, 2)

  def test_extract_url_with_http(self):
    text = "Check out this website: http://www.example.com for more info."
    expected_url = "http://www.example.com"
    self.assertEqual(
        ad_copy_generator.extract_url_from_string(text), expected_url
    )

  def test_extract_url_with_https(self):
    text = "Secure site: https://www.google.com"
    expected_url = "https://www.google.com"
    self.assertEqual(
        ad_copy_generator.extract_url_from_string(text), expected_url
    )

  def test_extract_url_with_complex_path(self):
    text = "API endpoint: https://api.example.com/v1/users/1234?key=abc"
    expected_url = "https://api.example.com/v1/users/1234?key=abc"
    self.assertEqual(
        ad_copy_generator.extract_url_from_string(text), expected_url
    )

  def test_no_url_present(self):
    text = "No URL in this text."
    self.assertFalse(
        ad_copy_generator.extract_url_from_string(text)
    )  # Check for False

  def test_multiple_urls(self):
    # This function is designed to extract the first URL it finds
    text = "Visit http://www.example.com or https://www.google.com"
    expected_url = "http://www.example.com"
    self.assertEqual(
        ad_copy_generator.extract_url_from_string(text), expected_url
    )

  def test_valid_urls(self):
    valid_urls = [
        "http://www.example.com",
        "https://www.google.com",
        "https://subdomain.example.co.uk/some/path",
        "https://google.com/path/with-hyphens",
    ]
    for url in valid_urls:
      self.assertTrue(ad_copy_generator.is_valid_url(url))

  def test_invalid_urls(self):
    invalid_urls = [
        "www.example.com",  # Missing protocol
        "htt://www.example.com",  # Invalid protocol
        "https://.com",  # Missing domain name
        "https://example",  # Missing TLD
        "ftp://example.com",  # Unsupported protocol
        "https://example.com/ path/with/spaces",  # Spaces in the path
    ]
    for url in invalid_urls:
      self.assertFalse(ad_copy_generator.is_valid_url(url))

  @mock.patch("bs4.BeautifulSoup")
  @mock.patch("requests.get")
  def test_extract_urls_for_keyword_instructions(
      self,
      mock_requests_get,
      mock_beautiful_soup,
  ):

    mock_requests_get.return_value.raise_for_status.return_value = None
    mock_requests_get.return_value.content = (
        "<html><body><h1>Test Page</h1></body></html>"
    )
    mock_beautiful_soup.return_value.get_text.return_value = "Test Page"

    mock_requests_get.side_effect = [
        mock.DEFAULT,
        mock.DEFAULT,
        requests.exceptions.RequestException("Simulated Error"),
    ]

    keyword_instructions = [
        "This is some text without a URL.",
        "Visit https://www.example.com for more details.",
        "https://www.google.com",
        "invalid_url",
        "https://www.error.com",
    ]
    expected_output = [
        "This is some text without a URL.",
        (
            "Visit https://www.example.com for more details. ## Content of"
            " https://www.example.com: Test Page"
        ),
        "Web page content of https://www.google.com: Test Page",
        "invalid_url",
        "https://www.error.com",
    ]

    result = ad_copy_generator.extract_urls_for_keyword_instructions(
        keyword_instructions
    )
    self.assertEqual(len(result), len(expected_output))
    self.assertEqual(result, expected_output)


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mesop as me

from copycat import copycat
from copycat.ui import components
from copycat.ui import event_handlers
from copycat.ui import states


def advertiser_info_section(params: states.CopycatParamsState) -> None:
  """Renders the advertiser info section of the new instance page."""
  me.text(
      "Advertising Info:", style=me.Style(margin=me.Margin(bottom=15))
  )
  with components.row():
    me.input(
        label="Company Name",
        key="company_name",
        on_blur=event_handlers.update_copycat_parameter,
        value=params.company_name,
        appearance="outline",
        style=me.Style(margin=me.Margin.symmetric(horizontal=5)),
    )

    me.input(
        label="Ad Copy Language",
        on_blur=event_handlers.language_on_blur,
        value=params.language,
        appearance="outline",
        style=me.Style(margin=me.Margin.symmetric(horizontal=5)),
    )

  me.text("Ad Format:")
  me.radio(
      on_change=event_handlers.ad_format_on_change,
      options=[
          me.RadioOption(
              label="Responsive Search Ads", value="responsive_search_ad"
          ),
          me.RadioOption(label="Text Ads", value="text_ad"),
          me.RadioOption(label="Custom", value="custom"),
      ],
      value=params.ad_format,
      style=me.Style(margin=me.Margin(bottom=15)),
  )

  ad_format_is_not_custom = params.ad_format != "custom"
  with components.row():
    me.input(
        label="Max Headlines",
        key="max_headlines",
        on_blur=event_handlers.update_copycat_parameter,
        value=str(params.max_headlines),
        type="number",
        appearance="outline",
        disabled=ad_format_is_not_custom,
        style=me.Style(margin=me.Margin.symmetric(horizontal=5)),
    )

    me.input(
        label="Max Descriptions",
        key="max_descriptions",
        on_blur=event_handlers.update_copycat_parameter,
        value=str(params.max_descriptions),
        type="number",
        appearance="outline",
        disabled=ad_format_is_not_custom,
        style=me.Style(margin=me.Margin.symmetric(horizontal=5)),
    )

  with me.tooltip(
      message=(
          "Special variables are things like Dynamic Keyword Insertion"
          " (DKI) or Customizers. If you don't want Copycat to use these"
          " features in your generated ads, then you should replace them"
          " with their default values. If you do want Copycat to use them"
          " then keep them in but make sure the Style Guide clearly"
          " explains how they should be used."
      )
  ):
    me.text("How to handle special variables:")
    me.radio(
        on_change=event_handlers.update_copycat_parameter,
        key="how_to_handle_special_variables",
        options=[
            me.RadioOption(
                label="Replace with default value", value="replace"
            ),
            me.RadioOption(label="Keep", value="keep"),
        ],
        value=params.how_to_handle_special_variables,
        style=me.Style(margin=me.Margin(bottom=15)),
    )

  me.text("How to handle invalid training ads:")
  me.radio(
      on_change=event_handlers.update_copycat_parameter,
      key="on_invalid_ad",
      options=[
          me.RadioOption(label="Drop", value="drop"),
          me.RadioOption(label="Skip", value="skip"),
          me.RadioOption(label="Raise", value="raise"),
      ],
      value=params.on_invalid_ad,
      style=me.Style(margin=me.Margin(bottom=15)),
  )


def copycat_creation_settings_section(params: states.CopycatParamsState) -> None:
  """Renders the Copycat creation settings section of the new instance page."""
  me.text("Embedding model:", style=me.Style(margin=me.Margin(bottom=15)))
  with components.row():
    with me.tooltip(
        message="Dimensionality of embeddings must be between 10 and 768."
    ):
      me.input(
          label="Dimensionality",
          on_blur=event_handlers.embedding_model_dimensionality_on_blur,
          value=str(params.embedding_model_dimensionality),
          type="number",
          appearance="outline",
          style=me.Style(margin=me.Margin.symmetric(horizontal=5)),
      )
    with me.tooltip(message="Batch size must be between 1 and 250."):
      me.input(
          label="Batch Size",
          key="embedding_model_batch_size",
          on_blur=event_handlers.update_copycat_parameter,
          value=str(params.embedding_model_batch_size),
          type="number",
          appearance="outline",
          style=me.Style(margin=me.Margin.symmetric(horizontal=5)),
      )

  me.text("Exemplar selection method:")
  with me.tooltip(
      message=(
          "Exemplars are selected either randomly, or using Affinity"
          " Propogation to find clusters of ads and select the ads that"
          " are the most representative of each cluster. We recommend"
          " Affinity Propagation."
      )
  ):
    me.radio(
        on_change=event_handlers.update_copycat_parameter,
        key="exemplar_selection_method",
        options=[
            me.RadioOption(
                label=v.value.replace("_", " ").title(), value=v.value
            )
            for v in copycat.ExemplarSelectionMethod
        ],
        value=params.exemplar_selection_method,
        style=me.Style(margin=me.Margin(bottom=15)),
    )

  with components.row():
    me.input(
        label="Max Exemplar Ads",
        key="max_exemplar_ads",
        on_blur=event_handlers.update_copycat_parameter,
        value=str(params.max_exemplar_ads),
        type="number",
        appearance="outline",
        style=me.Style(margin=me.Margin.symmetric(horizontal=5)),
    )
    if (
        params.exemplar_selection_method
        == copycat.ExemplarSelectionMethod.AFFINITY_PROPAGATION.value
    ):
      me.input(
          label="Max Initial Ads",
          key="max_initial_ads",
          on_blur=event_handlers.update_copycat_parameter,
          value=str(params.max_initial_ads),
          type="number",
          appearance="outline",
          style=me.Style(margin=me.Margin.symmetric(horizontal=5)),
      )
  if (
      params.exemplar_selection_method
      == copycat.ExemplarSelectionMethod.AFFINITY_PROPAGATION.value
  ):
    me.text(
        "Use Custom Affinity Preference:",
        style=me.Style(margin=me.Margin(bottom=15)),
    )
    with components.row():
      with me.box(
          style=me.Style(
              margin=me.Margin.symmetric(vertical=10),
              align_items="center",
          )
      ):
        me.slide_toggle(
            key="use_custom_affinity_preference",
            on_change=event_handlers.update_copycat_parameter_from_slide_toggle,
            checked=params.use_custom_affinity_preference,
        )
      if params.use_custom_affinity_preference:
        me.input(
            label="Affinity Preference",
            key="custom_affinity_preference",
            on_blur=event_handlers.update_copycat_parameter,
            value=str(params.custom_affinity_preference),
            type="number",
            appearance="outline",
            style=me.Style(
                margin=me.Margin.symmetric(horizontal=5), width=100
            ),
        )


def new_instance() -> None:
  """Renders the new instance page."""
  params = me.state(states.CopycatParamsState)
  state = me.state(states.AppState)

  with components.rounded_box_section("Create Copycat Instance"):
    with components.row(width="100%"):
      with components.column(width="50%", gap=0):
        advertiser_info_section(params)

      with components.column(width="50%", gap=0):
        copycat_creation_settings_section(params)

    with components.conditional_tooltip(
        disabled=not state.google_sheet_is_valid,
        disabled_tooltip="Must first validate the google sheet.",
    ):
      with components.row(align_items="center"):
        me.button(
            label="Re-build Copycat Instance"
            if state.has_copycat_instance
            else "Build Copycat Instance",
            type="flat",
            on_click=event_handlers.build_new_copycat_instance,
            style=me.Style(margin=me.Margin.symmetric(vertical=10)),
        )
        if state.has_copycat_instance:
          me.icon("check_circle")
          me.text("Copycat Instance Exists")



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The main module for Copycat.

Contains the code to generate copycat ad copies.
"""

from collections.abc import Mapping
import dataclasses
import json
import logging
from typing import Any, Callable
import warnings

from vertexai import generative_models
import pandas as pd
import pydantic

from copycat import ad_copy_evaluator
from copycat import ad_copy_generator
from copycat import google_ads
from copycat import keyword_organiser
from copycat import style_guide as style_guide_generator

GoogleAd = google_ads.GoogleAd
GoogleAdFormat = google_ads.GoogleAdFormat
ValidationError = pydantic.ValidationError
ModelName = ad_copy_generator.ModelName
EmbeddingModelName = ad_copy_generator.EmbeddingModelName
TextGenerationRequest = ad_copy_generator.TextGenerationRequest
ExemplarSelectionMethod = ad_copy_generator.ExemplarSelectionMethod
EvaluationResults = ad_copy_evaluator.EvaluationResults
StyleGuideGenerator = style_guide_generator.StyleGuideGenerator
BirchAgglomerativeKeywordClusterer = keyword_organiser.BirchAgglomerativeKeywordClusterer

# Below are not used in this file, they are included for the user to easily
# adjust the safety settings in copycat without having to import
# generative_models from vertex ai.
HarmCategory = ad_copy_generator.generative_models.HarmCategory
HarmBlockThreshold = ad_copy_generator.generative_models.HarmBlockThreshold
ALL_SAFETY_SETTINGS_OFF = {
    harm_category: HarmBlockThreshold.BLOCK_NONE
    for harm_category in HarmCategory
}

ALL_SAFETY_SETTINGS_ONLY_HIGH = {
    harm_category: HarmBlockThreshold.BLOCK_ONLY_HIGH
    for harm_category in HarmCategory
}

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)


DEFAULT_SYSTEM_INSTRUCTION = """\
You are an expert marketing professional, working for {company_name}. You are
tasked with writing new headlines and descriptions for Google Ads in {language},
given a new set of keywords, that will maximize engagement and clicks.
Keywords are words or phrases that are used to match ads with the terms that
people are searching for, so the copy should be engaging for someone searching
for those keywords. For each ad you must produce a list of headlines and a list
of descriptions, and those headlines and descriptions should be varied, while 
making sense in any combination together - Google Ads will select different 
combinations of headlines and descriptions to display to users. Each headline 
must be no longer than {max_headline_length} characters, and each description 
must be no longer than {max_description_length} characters.
""".replace("\n", " ").replace("  ", " ")

COPYCAT_PARAMS_FILE_NAME = "copycat_params.json"


class CopycatResponseError(ValueError):
  """The error raised when the CopycatResponse is not successful."""


class CopycatResponse(pydantic.BaseModel):
  """The response from Copycat.

  Attributes:
    google_ad: The generated ad.
    keywords: The keywords used to generate the ad, as a comma separated string.
    evaluation_results: The evaluation results of the ad, including whether it
      is memorised from the training data, and it's style and keyword similarity
      metrics.
    success: Whether the generation was successful.
    error_message: The error message if the generation was not successful.
  """

  google_ad: GoogleAd
  keywords: str
  evaluation_results: ad_copy_evaluator.EvaluationResults

  @property
  def success(self) -> bool:
    return not self.error_message

  @property
  def error_message(self) -> str:
    return "\n".join(
        map(lambda x: f"- {x}", sorted(self.evaluation_results.errors))
    )

  @property
  def warning_message(self) -> str:
    return "\n".join(
        map(lambda x: f"- {x}", sorted(self.evaluation_results.warnings))
    )

  def raise_if_not_success(self) -> None:
    if not self.success:
      LOGGER.error("CopycatResponse is not successful: %s", self.error_message)
      raise CopycatResponseError(self.error_message)


@dataclasses.dataclass
class Copycat:
  """The Copycat model which generates ad copies in the advertisers style.

  Attributes:
    ad_copy_vectorstore: The vectorstore containing the training ad copies.
    ad_format: The ad format that copycat will generate (same as the ad format
      of the examples in the vectorstore).
    output_parser: The output parser to use to parse the output of the chat
      model to a GoogleAd.
    ad_copy_evaluator: The ad copy evaluator to use to evaluate the generated ad
      copies.
    style_guide: The style guide to use to generate the ad copies.
  """

  ad_copy_vectorstore: ad_copy_generator.AdCopyVectorstore
  ad_format: GoogleAdFormat
  style_guide: str = ""

  @property
  def ad_copy_evaluator(self) -> ad_copy_evaluator.AdCopyEvaluator:
    return ad_copy_evaluator.AdCopyEvaluator(
        self.ad_format,
        ad_copy_vectorstore=self.ad_copy_vectorstore,
    )

  @classmethod
  def _clean_invalid_ads(
      cls,
      data: pd.DataFrame,
      ad_format: GoogleAdFormat,
      on_invalid_ad: str,
      replace_special_variables_with_default: bool,
  ) -> pd.DataFrame:
    """Cleans the invalid ads from the training data.

    Args:
      data: The training data containing the headlines, descriptions and
        keywords.
      ad_format: The ad format used in this vectorstore.
      on_invalid_ad: The action to take on invalid ads. Must be one of "raise",
        "skip", or "drop".
      replace_special_variables_with_default: Whether to replace Google Ads
        special variables with their default values.

    Returns:
      The training data with the invalid ads handled. If on_invalid_ad is
      "raise", then an error is raised. If on_invalid_ad is "skip", then the
      invalid ads are kept in the training data. If on_invalid_ad is "drop",
      then the invalid ads are dropped from the training data.

    Raises:
      ValueError: If on_invalid_ad is not one of "raise", "skip", or "drop".
      ValueError: If there are invalid ads in the training data and
        on_invalid_ad is "raise".
    """
    evaluator = ad_copy_evaluator.AdCopyEvaluator(ad_format)
    LOGGER.info(
        "Cleaning invalid ads from the training data. Will %s invalid ads.",
        on_invalid_ad,
    )

    if on_invalid_ad not in ["raise", "skip", "drop"]:
      LOGGER.error("Invalid value for on_invalid_ad: %s", on_invalid_ad)
      raise ValueError(
          f"Invalid value for on_invalid_ad: {on_invalid_ad}. Must be one of"
          " 'raise', 'skip', or 'drop'."
      )

    if replace_special_variables_with_default:
      data["headlines"] = data["headlines"].apply(
          lambda headlines: list(
              map(google_ads.parse_google_ads_special_variables, headlines)
          )
      )
      data["descriptions"] = data["descriptions"].apply(
          lambda descriptions: list(
              map(google_ads.parse_google_ads_special_variables, descriptions)
          )
      )
      is_invalid = data.apply(
          lambda row: not evaluator.is_valid(
              GoogleAd(
                  headlines=row["headlines"], descriptions=row["descriptions"]
              )
          )
          or evaluator.has_unfillable_google_ads_special_variables(
              GoogleAd(
                  headlines=row["headlines"], descriptions=row["descriptions"]
              )
          ),
          axis=1,
      )
    else:
      is_invalid = data.apply(
          lambda row: not evaluator.is_valid(
              GoogleAd(
                  headlines=row["headlines"], descriptions=row["descriptions"]
              )
          ),
          axis=1,
      )
    n_invalid_ads = is_invalid.sum()
    frac_invalid_ads = n_invalid_ads / len(data)
    error_message = (
        f"{n_invalid_ads:,} ({frac_invalid_ads:.2%}) invalid ads found in the"
        " training data."
    )

    if n_invalid_ads > 0:
      if on_invalid_ad == "raise":
        LOGGER.error(error_message)
        raise ValueError(error_message)
      elif on_invalid_ad == "skip":
        LOGGER.warning("%s Keeping them in the training data.", error_message)
        warnings.warn(error_message + " Keeping them in the training data.")
      elif on_invalid_ad == "drop":
        LOGGER.warning(
            "%s Dropping them from the training data.", error_message
        )
        warnings.warn(error_message + " Dropping them from the training data.")
        data = data[~is_invalid]
    else:
      LOGGER.info("No invalid ads found in the training data.")

    return data

  @classmethod
  def create_from_pandas(
      cls,
      *,
      training_data: pd.DataFrame,
      embedding_model_name: str | EmbeddingModelName,
      ad_format: str | GoogleAdFormat,
      on_invalid_ad: str = "drop",
      embedding_model_dimensionality: int = 256,
      vectorstore_max_initial_ads: int = 2000,
      vectorstore_max_exemplar_ads: int = 200,
      vectorstore_affinity_preference: float | None = None,
      vectorstore_exemplar_selection_method: (
          str | ExemplarSelectionMethod
      ) = "affinity_propagation",
      embedding_model_batch_size: int = 50,
      replace_special_variables_with_default: bool = False,
  ) -> "Copycat":
    """Creates a Copycat model from a pandas dataframe.

    The pandas dataframe must contain the columns "headline", "description", and
    "keywords", with a different row per ad.

    Args:
      training_data: The historical ad copies to learn the style from. Must
        contain the columns "headline", "description", and "keywords".
      embedding_model_name: The name of the embedding model to use to create the
        ad copy vectorstore.
      ad_format: The ad format that copycat will generate (same as the ad format
        of the examples in the training data).
      on_invalid_ad: How to handle invalid ads in the training data. Must be one
        of "drop", "raise", or "skip". "drop" means that the invalid ads will be
        dropped. "raise" means that an exception will be raised. "skip" means
        that the invalid ads will remain in the training data.
      embedding_model_dimensionality: The dimensionality of the embedding model.
      vectorstore_max_initial_ads: The maximum number of ads to use from the
        training data when creating the ad copy vectorstore.
      vectorstore_max_exemplar_ads: The maximum number of exemplar ads to use in
        the ad copy vectorstore.
      vectorstore_affinity_preference: The affinity preference to use when
        finding exemplar ads.
      vectorstore_exemplar_selection_method: The method to use to select the
        exemplar ads. Either "affinity_propagation" or "random". Defaults to
        "affinity_propagation".
      embedding_model_batch_size: The batch size to use when generating
        embeddings.
      replace_special_variables_with_default: Whether to replace Google Ads
        special variables with their default values. These are things like
        Dynamic Keyword Insertion and Customizers. If you replace them with
        their default values then Copycat won't try to learn them, and it will
        just generate generic ads without DKI or Customizers. If you keep them
        then it might try to generate them, but this can sometimes not work
        well.

    Returns:
      A Copycat model.

    Raises:
      ValueError: If the training data does not contain the required columns.
    """
    training_data = training_data.copy()

    if isinstance(ad_format, str):
      ad_format = google_ads.get_google_ad_format(ad_format)

    required_columns = {"headlines", "descriptions", "keywords"}
    missing_columns = required_columns - set(training_data.columns)
    if missing_columns:
      LOGGER.error(
          "Training data must contain the columns %s. Missing columns: %s.",
          sorted(required_columns),
          sorted(missing_columns),
      )
      raise ValueError(
          f"Training data must contain the columns {sorted(required_columns)}."
          f" Missing columns: {sorted(missing_columns)}."
      )

    training_data = cls._clean_invalid_ads(
        training_data,
        ad_format,
        on_invalid_ad,
        replace_special_variables_with_default,
    )

    ad_copy_vectorstore = (
        ad_copy_generator.AdCopyVectorstore.create_from_pandas(
            training_data=training_data,
            embedding_model_name=embedding_model_name,
            dimensionality=embedding_model_dimensionality,
            max_initial_ads=vectorstore_max_initial_ads,
            max_exemplar_ads=vectorstore_max_exemplar_ads,
            affinity_preference=vectorstore_affinity_preference,
            embeddings_batch_size=embedding_model_batch_size,
            exemplar_selection_method=vectorstore_exemplar_selection_method,
        )
    )

    return cls(
        ad_copy_vectorstore=ad_copy_vectorstore,
        ad_format=ad_format,
    )

  @classmethod
  def from_dict(cls, params: dict[str, Any]) -> "Copycat":
    """Loads the model from the provided dict.

    Schema of params:
      ad_copy_vectorstore: The ad copy vectorstore to use. This is a dict
        containing the parameters to use to create the ad copy vectorstore. See
        AdCopyVectorstore.from_dict for the required keys.
      ad_format_params: The parameters to use to create the ad format. Must
        include the following keys:
        - max_headlines: The maximum number of headlines to generate.
        - max_descriptions: The maximum number of descriptions to generate.

    Args:
      params: The dict containing the parameters to use to create the Copycat
        model. See above for the schema.

    Returns:
      A Copycat model.

    Raises:
      KeyError: If any of the required keys are missing from the dict.
    """
    required_keys = {
        "ad_copy_vectorstore",
        "ad_format_params",
    }
    missing_keys = required_keys - set(params.keys())
    if missing_keys:
      LOGGER.error("Missing required keys: %s", missing_keys)
      raise KeyError(f"Missing required keys: {missing_keys}")

    return cls(
        ad_copy_vectorstore=ad_copy_generator.AdCopyVectorstore.from_dict(
            params["ad_copy_vectorstore"]
        ),
        ad_format=GoogleAdFormat(**params["ad_format_params"]),
        style_guide=params.get("style_guide", ""),
    )

  def to_dict(self) -> dict[str, Any]:
    """Serializes the model to a dict."""
    return {
        "ad_format_params": self.ad_format.model_dump(),
        "ad_copy_vectorstore": self.ad_copy_vectorstore.to_dict(),
        "style_guide": self.style_guide,
    }

  @classmethod
  def from_json(cls, json_string: str) -> "Copycat":
    """Loads the model from the provided json string.

    Schema of params:
      ad_copy_vectorstore: The ad copy vectorstore to use. This is a dict
        containing the parameters to use to create the ad copy vectorstore. See
        AdCopyVectorstore.from_dict for the required keys.
      ad_format_params: The parameters to use to create the ad format. Must
        include the following keys:
        - max_headlines: The maximum number of headlines to generate.
        - max_descriptions: The maximum number of descriptions to generate.

    Args:
      json_string: The json string containing the parameters to use to create
        the Copycat model. See above for the schema.

    Returns:
      A Copycat model.

    Raises:
      KeyError: If any of the required keys are missing from the dict.
    """
    return cls.from_dict(json.loads(json_string))

  def to_json(self) -> str:
    """Serializes the model to a json string."""
    return json.dumps(self.to_dict())

  def construct_responses(
      self,
      raw_generated_ads: list[generative_models.Candidate],
      keywords: list[str],
      existing_ad_copies: list[GoogleAd],
  ) -> list[CopycatResponse]:
    """Constructs a CopycatResponse from a generated GoogleAd.

    Args:
      raw_generated_ads: The unprocessed generated ads as a generation
        candidates.
      keywords: The keywords used to generate the ads.
      existing_ad_copies: The existing ad copies if they exist, to be merged
        with the generated ad copies.

    Returns:
      A CopycatResponse object.
    """
    empty_evaluation_results = ad_copy_evaluator.EvaluationResults(
        errors=[],
        warnings=[],
        headlines_are_memorised=None,
        descriptions_are_memorised=None,
        keyword_similarity=None,
        style_similarity=None,
    )

    responses = []
    for keywords_i, raw_generated_ad_i, existing_ad_copy_i in zip(
        keywords, raw_generated_ads, existing_ad_copies
    ):
      if (
          raw_generated_ad_i.finish_reason
          is not ad_copy_generator.FinishReason.STOP
      ):
        responses.append(
            CopycatResponse(
                google_ad=existing_ad_copy_i.model_copy(),
                keywords=keywords_i,
                evaluation_results=empty_evaluation_results.model_copy(
                    update=dict(errors=[str(raw_generated_ad_i)])
                ),
            )
        )
        LOGGER.error(
            "Generated ad did not finish. Complete response: %s.",
            raw_generated_ad_i,
        )
        continue

      try:
        LOGGER.debug(
            "Generation for keywords %s\n\n%s",
            keywords_i,
            raw_generated_ad_i.content.parts[0].text,
        )
        generated_ad_copy = GoogleAd.model_validate_json(
            raw_generated_ad_i.content.parts[0].text
        )
      except ValidationError as e:
        responses.append(
            CopycatResponse(
                google_ad=existing_ad_copy_i.model_copy(),
                keywords=keywords_i,
                evaluation_results=empty_evaluation_results.model_copy(
                    update=dict(errors=[str(e)])
                ),
            )
        )
        LOGGER.error(
            "Generated ad was not matching the expected json format."
            " Error: %s.",
            e,
        )
        continue

      ad_copy = existing_ad_copy_i.model_copy() + generated_ad_copy
      ad_copy_generator.remove_invalid_headlines_and_descriptions(
          ad_copy, self.ad_format
      )

      responses.append(
          CopycatResponse(
              google_ad=ad_copy,
              keywords=keywords_i,
              evaluation_results=empty_evaluation_results,
          )
      )

    return responses

  def _evaluate_responses(
      self,
      responses: list[CopycatResponse],
      allow_memorised_headlines: bool,
      allow_memorised_descriptions: bool,
  ) -> list[CopycatResponse]:
    """Evaluates the responses if the ad copy is not empty.

    If the ad copy is empty, then it is not evaluated.

    Args:
      responses: The responses to evaluate.
      allow_memorised_headlines: Whether to allow memorised headlines.
      allow_memorised_descriptions: Whether to allow memorised descriptions.

    Returns:
      The evaluated responses.
    """
    evaluation_results_list = self.ad_copy_evaluator.evaluate_batch(
        ad_copies=[response.google_ad for response in responses],
        allow_memorised_headlines=allow_memorised_headlines,
        allow_memorised_descriptions=allow_memorised_descriptions,
        keywords=[response.keywords for response in responses],
    )
    evaluated_responses = []
    for response, evaluation_results in zip(responses, evaluation_results_list):
      if self.ad_copy_evaluator.is_empty(response.google_ad):
        evaluated_responses.append(response.model_copy())
      else:
        merged_evaluation_results = evaluation_results.model_copy(
            update=dict(
                warnings=response.evaluation_results.warnings
                + evaluation_results.warnings,
                errors=response.evaluation_results.errors
                + evaluation_results.errors,
            )
        )
        evaluated_responses.append(
            response.model_copy(
                update=dict(evaluation_results=merged_evaluation_results)
            )
        )
    return evaluated_responses

  def construct_text_generation_requests_for_new_ad_copy(
      self,
      *,
      keywords: list[str],
      keywords_specific_instructions: list[str] | None = None,
      style_guide: str | None = None,
      system_instruction: str = DEFAULT_SYSTEM_INSTRUCTION,
      num_in_context_examples: int = 10,
      model_name: ModelName | str = ModelName.GEMINI_1_5_FLASH,
      temperature: float = 0.95,
      top_k: int = 20,
      top_p: float = 0.95,
      safety_settings: ad_copy_generator.SafetySettingsType | None = None,
      system_instruction_kwargs: dict[str, Any] | None = None,
      existing_headlines: list[list[str]] | None = None,
      existing_descriptions: list[list[str]] | None = None,
  ) -> list[TextGenerationRequest]:
    """Constructs a request for generating a new ad copy.

    This prompt consists of a system prompt, a style guide, and a number of
    in context examples. The in context examples are retrieved from the ad copy
    vectorstore.

    If there are already some headlines or descriptions, then the prompt will
    ask Copycat to extend the ad copy to the required number of headlines and
    descriptions for the ad format.

    Args:
      keywords: The list of keywords to use to generate the ad copies. This
        should be a list of strings, where each string is a comma separated list
        of keywords.
      keywords_specific_instructions: The list of keywords specific instructions
        to use. Defaults to a list of empty strings.
      style_guide: The style guide to use. If None, then the style guide from
        the Copycat model will be used.
      system_instruction: The system instruction to use.
      num_in_context_examples: The number of in context examples to use.
      model_name: The name of the gemini model to use.
      temperature: The temperature to use for the chat model.
      top_k: The top-k to use for the chat model.
      top_p: The top-p to use for the chat model.
      safety_settings: The safety settings for the chat model.
      system_instruction_kwargs: Additional arguments to pass to the system
        instruction.
      existing_headlines: The existing headlines for the ad copy. If no
        headlines then pass None.
      existing_descriptions: The existing descriptions for the ad copy. If no
        descriptions then pass None.

    Returns:
      A text generation request, containing the prompt, system instruction, and
      model parameters.
    """
    if keywords_specific_instructions is None:
      keywords_specific_instructions = [""] * len(keywords)
    if existing_headlines is None:
      existing_headlines = [None] * len(keywords)
    if existing_descriptions is None:
      existing_descriptions = [None] * len(keywords)

    system_instruction_kwargs = system_instruction_kwargs or {}
    default_system_instruction_kwargs = {
        "max_headline_length": self.ad_format.max_headline_length,
        "max_description_length": self.ad_format.max_description_length,
    }
    system_instruction_kwargs = (
        default_system_instruction_kwargs | system_instruction_kwargs
    )
    if style_guide is None:
      style_guide = self.style_guide
    system_instruction = ad_copy_generator.construct_system_instruction(
        system_instruction=system_instruction,
        style_guide=style_guide,
        system_instruction_kwargs=system_instruction_kwargs,
    )

    relavent_example_ads = self.ad_copy_vectorstore.get_relevant_ads(
        keywords,
        k=num_in_context_examples,
    )

    existing_ad_copies = [
        GoogleAd(
            headlines=headlines_i or [],
            descriptions=descriptions_i or [],
        )
        for headlines_i, descriptions_i in zip(
            existing_headlines,
            existing_descriptions,
        )
    ]

    prompts = [
        ad_copy_generator.construct_new_ad_copy_prompt(
            example_ads=relevant_example_ads_i,
            keywords=keywords_i,
            ad_format=self.ad_format,
            existing_ad_copy=existing_ad_copy_i,
            keywords_specific_instructions=keywords_specific_instructions_i,
        )
        for keywords_i, keywords_specific_instructions_i, relevant_example_ads_i, existing_ad_copy_i in zip(
            keywords,
            keywords_specific_instructions,
            relavent_example_ads,
            existing_ad_copies,
        )
    ]

    requests = [
        TextGenerationRequest(
            keywords=keywords_i,
            existing_ad_copy=existing_ad_copy_i,
            prompt=prompt_i,
            system_instruction=system_instruction,
            chat_model_name=ModelName(model_name),
            temperature=temperature,
            top_k=top_k,
            top_p=top_p,
            safety_settings=safety_settings,
        )
        for keywords_i, prompt_i, existing_ad_copy_i in zip(
            keywords, prompts, existing_ad_copies
        )
    ]

    return requests

  def _generate_new_ad_copy_from_requests(
      self,
      requests: list[TextGenerationRequest],
  ) -> list[CopycatResponse]:
    """Generates a new ad copy from a list of requests.

    Args:
      requests: The requests to generate the ad copy from.

    Returns:
      A list of CopycatResponses.
    """
    generations = [
        response.candidates[0]
        for response in ad_copy_generator.generate_google_ad_json_batch(
            requests
        )
    ]
    keywords = [request.keywords for request in requests]
    existing_ad_copies = [request.existing_ad_copy for request in requests]

    responses = self.construct_responses(
        generations,
        keywords,
        existing_ad_copies,
    )
    return responses

  def generate_new_ad_copy(
      self,
      *,
      keywords: list[str],
      keywords_specific_instructions: list[str] | None = None,
      style_guide: str | None = None,
      system_instruction: str = DEFAULT_SYSTEM_INSTRUCTION,
      num_in_context_examples: int = 10,
      model_name: ModelName | str = ModelName.GEMINI_1_5_FLASH,
      temperature: float = 0.95,
      top_k: int = 20,
      top_p: float = 0.95,
      allow_memorised_headlines: bool = True,
      allow_memorised_descriptions: bool = False,
      safety_settings: ad_copy_generator.SafetySettingsType | None = None,
      system_instruction_kwargs: dict[str, Any] | None = None,
      existing_headlines: list[list[str]] | None = None,
      existing_descriptions: list[list[str]] | None = None,
  ) -> list[CopycatResponse]:
    """Generates a new ad copy.

    Args:
      keywords: The list of keywords to use to generate the ad copies. This
        should be a list of strings, where each string is a comma separated list
        of keywords.
      keywords_specific_instructions: The list of keywords specific instructions
        to use. Defaults to a list of empty strings.
      style_guide: The style guide to use. If None, then the style guide from
        the Copycat model will be used.
      system_instruction: The system instruction to use.
      num_in_context_examples: The number of in context examples to use.
      model_name: The name of the chat model to use.
      temperature: The temperature to use for the chat model.
      top_k: The top-k to use for the chat model.
      top_p: The top-p to use for the chat model.
      allow_memorised_headlines: Whether to allow memorised headlines.
      allow_memorised_descriptions: Whether to allow memorised descriptions.
      safety_settings: The safety settings for the chat model.
      system_instruction_kwargs: Additional arguments to pass to the system
        instruction.
      existing_headlines: The existing headlines for the ad copy. If no
        headlines then pass None.
      existing_descriptions: The existing descriptions for the ad copy. If no
        descriptions then pass None.

    Returns:
      A CopycatResponse object.

    Raises:
      ValueError: If keywords, keywords_specific_instructions, existing
        headlines or existing descriptions have different lengths.
      RuntimeError: If the number of responses does not match the number of
        keywords. This shouldn't happen, if it happens it indicates a bug in the
        code.
    """
    if keywords_specific_instructions is None:
      keywords_specific_instructions = [""] * len(keywords)
    if existing_headlines is None:
      existing_headlines = [None] * len(keywords)
    if existing_descriptions is None:
      existing_descriptions = [None] * len(keywords)

    if len(keywords) != len(keywords_specific_instructions):
      LOGGER.error(
          "keywords and keywords_specific_instructions must have the same"
          " length."
      )
      raise ValueError(
          "keywords and keywords_specific_instructions must have the same"
          " length."
      )
    if len(existing_headlines) != len(keywords):
      LOGGER.error("keywords and existing_headlines must have the same length.")
      raise ValueError(
          "keywords and existing_headlines must have the same length."
      )
    if len(existing_descriptions) != len(keywords):
      LOGGER.error(
          "keywords and existing_descriptions must have the same length."
      )
      raise ValueError(
          "keywords and existing_descriptions must have the same length."
      )

    requests = self.construct_text_generation_requests_for_new_ad_copy(
        keywords=keywords,
        keywords_specific_instructions=keywords_specific_instructions,
        num_in_context_examples=num_in_context_examples,
        style_guide=style_guide,
        system_instruction=system_instruction,
        model_name=model_name,
        temperature=temperature,
        top_k=top_k,
        top_p=top_p,
        safety_settings=safety_settings,
        system_instruction_kwargs=system_instruction_kwargs,
        existing_headlines=existing_headlines,
        existing_descriptions=existing_descriptions,
    )

    responses = self._generate_new_ad_copy_from_requests(requests)

    evaluated_responses = self._evaluate_responses(
        responses,
        allow_memorised_headlines=allow_memorised_headlines,
        allow_memorised_descriptions=allow_memorised_descriptions,
    )

    if len(evaluated_responses) != len(keywords):
      LOGGER.error(
          "The number of responses does not match the number of keywords."
      )
      raise RuntimeError(
          "The number of responses does not match the number of keywords."
      )

    return evaluated_responses

  def generate_new_ad_copy_for_dataframe(
      self,
      data: pd.DataFrame,
      *,
      keywords_column: str = "keywords",
      keywords_specific_instructions_column: (
          str
      ) = "keywords_specific_instructions",
      existing_headlines_column: str = "existing_headlines",
      existing_descriptions_column: str = "existing_descriptions",
      **static_params: Any,
  ) -> pd.Series:
    """Applies the generate new ad copy function to a dataframe.

    This will generate a new ad copy for each row in the dataframe, using the
    keywords, keywords_specific_instructions, existing_headlines and
    existing_descriptions columns from that row. All other parameters are
    static and will be passed to the generate_new_ad_copy function.

    Args:
      data: The dataframe to apply the generate new ad copy function to.
      keywords_column: The name of the column containing the keywords.
      keywords_specific_instructions_column: The name of the column containing
        the keywords specific instructions. If the column is not present then
        None will be used as the value.
      existing_headlines_column: The name of the column containing the existing
        headlines. If the column is not present then None will be used as the
        value.
      existing_descriptions_column: The name of the column containing the
        existing descriptions. If the column is not present then None will be
        used as the value.
      **static_params: The static parameters to pass to the generate new ad copy
        function. See the generate_new_ad_copy function for the list of
        available parameters.

    Returns:
      A series of CopycatResponses.

    Raises:
      ValueError: If the dataframe does not contain the required keywords
        column.
    """
    if keywords_column in data.columns:
      keywords = data[keywords_column]
    else:
      LOGGER.error(
          "The dataframe does not contain the required column: %s",
          keywords_column,
      )
      raise ValueError(
          "The dataframe does not contain the required column:"
          f" {keywords_column}"
      )

    if keywords_specific_instructions_column in data.columns:
      keywords_specific_instructions = data[
          keywords_specific_instructions_column
      ]
    else:
      LOGGER.warning(
          "The dataframe does not contain the optional column: %s",
          keywords_specific_instructions_column,
      )
      keywords_specific_instructions = [None] * len(keywords)

    if existing_headlines_column in data.columns:
      existing_headlines = data[existing_headlines_column]
    else:
      LOGGER.warning(
          "The dataframe does not contain the optional column: %s",
          existing_headlines_column,
      )
      existing_headlines = [None] * len(keywords)

    if existing_descriptions_column in data.columns:
      existing_descriptions = data[existing_descriptions_column]
    else:
      LOGGER.warning(
          "The dataframe does not contain the optional column: %s",
          existing_descriptions_column,
      )
      existing_descriptions = [None] * len(keywords)

    generated_responses = self.generate_new_ad_copy(
        keywords=keywords,
        keywords_specific_instructions=keywords_specific_instructions,
        existing_headlines=existing_headlines,
        existing_descriptions=existing_descriptions,
        **static_params,
    )
    return pd.Series(generated_responses, index=data.index)

  def generate_style_guide(
      self,
      *,
      company_name: str,
      files_uri: str = "",
      additional_style_instructions: str = "",
      model_name: ModelName | str = ModelName.GEMINI_1_5_PRO,
      safety_settings: ad_copy_generator.SafetySettingsType | None = None,
      temperature: float = 0.95,
      top_k: int = 20,
      top_p: float = 0.95,
      use_exemplar_ads: bool = True,
  ) -> str:
    """Generates a style guide for the ad copy.

    The style guide is returned and also stored in the Copycat instance.

    Args:
      company_name: The name of the company.
      files_uri: The URI of the files to use to generate the style guide.
      additional_style_instructions: Additional instructions to use to generate
        the style guide.
      model_name: The name of the chat model to use.
      safety_settings: The safety settings for the chat model.
      temperature: The temperature to use for the chat model.
      top_k: The top-k to use for the chat model.
      top_p: The top-p to use for the chat model.
      use_exemplar_ads: Whether to use exemplar ads to generate the style guide.

    Returns:
      The generated style guide.

    Raises:
      ValueError: If the company name is not provided.
      ValueError: If the files URI is not provided and use_exemplar_ads is
        False.
      RuntimeError: If the style guide generation was not successful.
    """
    if not use_exemplar_ads and not files_uri:
      raise ValueError(
          "Must either provide a files URI or set use_exemplar_ads to True."
      )
    if not company_name:
      raise ValueError("Must provide a company name.")

    generator = StyleGuideGenerator()
    if files_uri:
      LOGGER.info("Checking for files in the GCP bucket %s", files_uri)
      generator.get_all_files(files_uri)

    model_response = generator.generate_style_guide(
        brand_name=company_name,
        ad_copy_vectorstore=self.ad_copy_vectorstore,
        additional_style_instructions=additional_style_instructions,
        model_name=model_name,
        safety_settings=safety_settings,
        temperature=temperature,
        top_k=top_k,
        top_p=top_p,
    ).candidates[0]
    if model_response.finish_reason is not ad_copy_generator.FinishReason.STOP:
      message = (
          "Style guide generation was not successful, complete response:"
          f" {model_response}"
      )
      LOGGER.error(message)
      raise RuntimeError(message)

    self.style_guide = model_response.content.text
    return self.style_guide



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest
from absl.testing import parameterized

from copycat import google_ads


class GoogleAdTest(parameterized.TestCase):

  def test_google_ad_can_be_converted_to_string_representation(self):
    google_ad = google_ads.GoogleAd(
        headlines=["Headline 1", "Headline 2"],
        descriptions=["Description 1.", "Description 2."],
    )
    self.assertEqual(
        str(google_ad),
        "**Headline 1 | Headline 2**\nDescription 1. Description 2.",
    )

  def test_google_ad_accepts_unicode_characters(self):
    google_ad = google_ads.GoogleAd(
        headlines=["Headline 1", "Headline 2"],
        descriptions=["Description 1.", "Description with unicode weiß."],
    )
    self.assertEqual(
        str(google_ad),
        "**Headline 1 | Headline 2**\nDescription 1. Description with unicode"
        " weiß.",
    )

  def test_google_ad_can_be_hashed_deterministically(self):
    # Must be hashable so it can be efficiently deduplicated.
    google_ad_1 = google_ads.GoogleAd(
        headlines=["Headline 1", "Headline 2"],
        descriptions=["Description 1.", "Description 2."],
    )
    google_ad_2 = google_ads.GoogleAd(
        headlines=["Headline 1", "Headline 2"],
        descriptions=["Description 1.", "Description 2."],
    )
    google_ad_3 = google_ads.GoogleAd(
        headlines=["Headline 2", "Headline 3"],
        descriptions=["Description 2.", "Description 3."],
    )

    self.assertEqual(hash(google_ad_1), hash(google_ad_2))
    self.assertNotEqual(hash(google_ad_1), hash(google_ad_3))

  @parameterized.parameters(
      ("responsive_search_ad", google_ads.RESPONSIVE_SEARCH_AD_FORMAT),
      ("text_ad", google_ads.TEXT_AD_FORMAT),
  )
  def test_get_google_ad_format_retuens_expected_format(
      self, name, expected_format
  ):
    self.assertEqual(google_ads.get_google_ad_format(name), expected_format)

  def test_get_google_ad_format_raises_error_for_invalid_format(self):
    with self.assertRaises(ValueError):
      google_ads.get_google_ad_format("invalid_format")

  @parameterized.parameters([
      ("No DKI is unchanged", "No DKI is unchanged"),
      (
          (
              "A few different DKIs: {KeyWord:all first letters capital}"
              " {Keyword:first letter first word capital} {keyword:all"
              " lowercase} {KEYWord:first word all caps} {KeyWORD:last word all"
              " caps} {CUSTOMIZER.something:my customizer}"
          ),
          (
              "A few different DKIs: All First Letters Capital First letter"
              " first word capital all lowercase FIRST Word All Caps Last Word"
              " All CAPS my customizer"
          ),
      ),
      (
          "Non DKI brackets are unchanged {not a keyword}",
          "Non DKI brackets are unchanged {not a keyword}",
      ),
  ])
  def test_parse_google_ads_special_variables(self, text, expected):
    got = google_ads.parse_google_ads_special_variables(text=text)
    self.assertEqual(expected, got)


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Setup for copycat package."""

import os

import setuptools

CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))


def _get_readme():
  try:
    readme = open(
        os.path.join(CURRENT_DIR, "../README.md"), encoding="utf-8"
    ).read()
  except OSError:
    readme = ""
  return readme


def _get_version():
  with open(os.path.join(CURRENT_DIR, "copycat", "__init__.py")) as fp:
    for line in fp:
      if line.startswith("__version__") and "=" in line:
        version = line[line.find("=") + 1 :].strip(" '\"\n")
        if version:
          return version
    raise ValueError("`__version__` not defined in `copycat/__init__.py`")


def _parse_requirements(path):
  with open(os.path.join(CURRENT_DIR, path)) as f:
    return [
        line.rstrip()
        for line in f
        if not (line.isspace() or line.startswith("#"))
    ]


VERSION = _get_version()
README = _get_readme()
INSTALL_REQUIREMENTS = _parse_requirements(
    os.path.join(CURRENT_DIR, "requirements.txt")
)
TEST_REQUIREMENTS = _parse_requirements(
    os.path.join(CURRENT_DIR, "requirements_tests.txt")
)
UI_REQUIREMENTS = _parse_requirements(
    os.path.join(CURRENT_DIR, "requirements_ui.txt")
)
GOOGLE_SHEETS_REQUIREMENTS = _parse_requirements(
    os.path.join(CURRENT_DIR, "requirements_google_sheets.txt")
)

setuptools.setup(
    name="gtech-copycat",
    version=VERSION,
    python_requires=">=3.10",
    description=(
        "Package for generating Google Search Ad Copies with AI that match the"
        " style of existing ads."
    ),
    long_description=README,
    long_description_content_type="text/markdown",
    author="Google gTech Ads",
    license="Apache 2.0",
    packages=setuptools.find_packages(),
    install_requires=INSTALL_REQUIREMENTS,
    tests_require=TEST_REQUIREMENTS,
    extras_require={
        "ui": UI_REQUIREMENTS,
        "google_sheets": GOOGLE_SHEETS_REQUIREMENTS,
    },
    url="https://github.com/google-marketing-solutions/copycat",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.10",
    ],
    include_package_data=True,
)
